# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang

import torch
import triton
import triton.language as tl

from fla.ops.utils import prepare_chunk_indices
from fla.ops.utils.op import gather
from fla.utils import IS_GATHER_SUPPORTED, USE_CUDA_GRAPH, autotune_cache_kwargs


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps)
        for num_warps in [1, 2, 4, 8, 16]
    ],
    key=['BT'],
    use_cuda_graph=USE_CUDA_GRAPH,
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def prepare_wy_repr_fwd_kernel_chunk32(
    A_ab,
    A_ab_inv,
    cu_seqlens,
    chunk_indices,
    T,
    H: tl.constexpr,
    BT: tl.constexpr,
    BC: tl.constexpr,  # placeholder, do not delete
    IS_VARLEN: tl.constexpr,
):
    i_t, i_bh = tl.program_id(0), tl.program_id(1)
    i_b, i_h = i_bh // H, i_bh % H
    if IS_VARLEN:
        i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T
    p_Aab = tl.make_block_ptr(A_ab + (bos*H + i_h) * BT, (T, BT), (H*BT, 1), (i_t * BT, 0), (BT, BT), (1, 0))
    p_Aab_inv = tl.make_block_ptr(A_ab_inv + (bos*H + i_h) * BT, (T, BT), (H*BT, 1), (i_t * BT, 0), (BT, BT), (1, 0))
    b_A_ab = tl.load(p_Aab, boundary_check=(0, 1))
    b_A_ab = tl.where(tl.arange(0, BT)[:, None] > tl.arange(0, BT)[None, :], b_A_ab, 0)
    for i in range(1, BT):
        mask = tl.arange(0, BT) == i
        b_a = tl.sum(tl.where(mask[:, None], b_A_ab, 0), 0)
        b_a = b_a + tl.sum(b_a[:, None] * b_A_ab, 0) * (tl.arange(0, BT) < i)
        b_A_ab = tl.where(mask[:, None], b_a, b_A_ab)
    b_A_ab += tl.arange(0, BT)[:, None] == tl.arange(0, BT)[None, :]
    tl.store(p_Aab_inv, b_A_ab.to(p_Aab_inv.dtype.element_ty), boundary_check=(0, 1))


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps, num_stages=num_stages)
        for num_warps in [2, 4, 8]
        for num_stages in [2, 3, 4]
    ],
    key=['BC'],
    use_cuda_graph=USE_CUDA_GRAPH,
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def prepare_wy_repr_fwd_kernel_chunk64(
    A_ab,
    A_ab_inv,
    cu_seqlens,
    chunk_indices,
    T,
    H: tl.constexpr,
    BT: tl.constexpr,
    BC: tl.constexpr,
    IS_VARLEN: tl.constexpr,
    GATHER_SUPPORTED: tl.constexpr = IS_GATHER_SUPPORTED,
):
    i_t, i_bh = tl.program_id(0), tl.program_id(1)
    i_b, i_h = i_bh // H, i_bh % H
    if IS_VARLEN:
        i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T

    p_A1 = tl.make_block_ptr(A_ab + (bos*H + i_h) * BT, (T, BT), (H*BT, 1), (i_t * BT, 0), (BC, BC), (1, 0))
    p_A2 = tl.make_block_ptr(A_ab + (bos*H + i_h) * BT, (T, BT), (H*BT, 1), (i_t * BT + BC, BC), (BC, BC), (1, 0))
    p_A3 = tl.make_block_ptr(A_ab + (bos*H + i_h) * BT, (T, BT), (H*BT, 1), (i_t * BT + BC, 0), (BC, BC), (1, 0))
    p_A_inv1 = tl.make_block_ptr(A_ab_inv + (bos*H + i_h) * BT, (T, BT), (H*BT, 1), (i_t * BT, 0), (BC, BC), (1, 0))
    p_A_inv2 = tl.make_block_ptr(A_ab_inv + (bos*H + i_h) * BT, (T, BT), (H*BT, 1), (i_t * BT + BC, BC), (BC, BC), (1, 0))
    p_A_inv3 = tl.make_block_ptr(A_ab_inv + (bos*H + i_h) * BT, (T, BT), (H*BT, 1), (i_t * BT + BC, 0), (BC, BC), (1, 0))
    p_A_inv4 = tl.make_block_ptr(A_ab_inv + (bos*H + i_h) * BT, (T, BT), (H*BT, 1), (i_t * BT, BC), (BC, BC), (1, 0))

    b_A = tl.load(p_A1, boundary_check=(0, 1))
    b_A2 = tl.load(p_A2, boundary_check=(0, 1))
    b_A3 = tl.load(p_A3, boundary_check=(0, 1))
    b_A = tl.where(tl.arange(0, BC)[:, None] > tl.arange(0, BC)[None, :], b_A, 0)
    b_A2 = tl.where(tl.arange(0, BC)[:, None] > tl.arange(0, BC)[None, :], b_A2, 0)

    for i in range(1, BC):
        if GATHER_SUPPORTED:
            row_idx = tl.full([1, BC], i, dtype=tl.int16)
            # [1, BK] -> [BK]
            b_a = tl.sum(gather(b_A, row_idx, axis=0), 0)
            b_a2 = tl.sum(gather(b_A2, row_idx, axis=0), 0)
        else:
            mask = tl.arange(0, BC) == i
            b_a = tl.sum(tl.where(mask[:, None], b_A, 0), 0)
            b_a2 = tl.sum(tl.where(mask[:, None], b_A2, 0), 0)
        mask = tl.arange(0, BC) == i
        # b_a = tl.sum(tl.where(mask[:, None], b_A, 0), 0)
        # b_a2 = tl.sum(tl.where(mask[:, None], b_A2, 0), 0)
        b_a = b_a + tl.sum(b_a[:, None] * b_A, 0) * (tl.arange(0, BC) < i)
        b_a2 = b_a2 + tl.sum(b_a2[:, None] * b_A2, 0) * (tl.arange(0, BC) < i)
        b_A = tl.where(mask[:, None], b_a, b_A)
        b_A2 = tl.where(mask[:, None], b_a2, b_A2)

    # blockwise computation of lower triangular matrix's inverse
    # i.e., [A11, 0; A21, A22]^-1 = [A11^-1, 0; -A22^-1 A21 A11^-1, A22^-1]
    b_A += tl.arange(0, BC)[:, None] == tl.arange(0, BC)[None, :]
    b_A2 += tl.arange(0, BC)[:, None] == tl.arange(0, BC)[None, :]
    b_A3 = tl.dot(tl.dot(b_A2, b_A3), b_A)
    # tl.debug_barrier()
    tl.store(p_A_inv1, b_A.to(p_A_inv1.dtype.element_ty, fp_downcast_rounding="rtne"), boundary_check=(0, 1))
    tl.store(p_A_inv2, b_A2.to(p_A_inv2.dtype.element_ty, fp_downcast_rounding="rtne"), boundary_check=(0, 1))
    tl.store(p_A_inv3, b_A3.to(p_A_inv3.dtype.element_ty, fp_downcast_rounding="rtne"), boundary_check=(0, 1))
    # causal mask
    tl.store(p_A_inv4, tl.zeros([BC, BC], dtype=tl.float32).to(p_A_inv4.dtype.element_ty), boundary_check=(0, 1))


@triton.heuristics({
    'IS_VARLEN': lambda args: args['cu_seqlens'] is not None,
})
@triton.autotune(
    configs=[
        triton.Config({}, num_warps=num_warps, num_stages=num_stages)
        for num_warps in [2, 4, 8, 16]
        for num_stages in [2, 3, 4]
    ],
    key=['H', 'K', 'V', 'BT', 'BK', 'BV', 'IS_VARLEN'],
    use_cuda_graph=USE_CUDA_GRAPH,
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def wu_fwd_kernel(
    w,
    u,
    ag,
    v,
    A_ab_inv,
    A_ak,
    cu_seqlens,
    chunk_indices,
    T,
    H: tl.constexpr,
    K: tl.constexpr,
    V: tl.constexpr,
    BT: tl.constexpr,
    BK: tl.constexpr,
    BV: tl.constexpr,
    IS_VARLEN: tl.constexpr,
):
    i_t, i_bh = tl.program_id(0), tl.program_id(1)
    i_b, i_h = i_bh // H, i_bh % H
    if IS_VARLEN:
        i_n, i_t = tl.load(chunk_indices + i_t * 2).to(tl.int32), tl.load(chunk_indices + i_t * 2 + 1).to(tl.int32)
        bos, eos = tl.load(cu_seqlens + i_n).to(tl.int32), tl.load(cu_seqlens + i_n + 1).to(tl.int32)
        T = eos - bos
    else:
        bos, eos = i_b * T, i_b * T + T
    o_s = tl.arange(0, BT)

    p_A_ab_inv = tl.make_block_ptr(A_ab_inv + (bos*H + i_h) * BT, (T, BT), (H*BT, 1), (i_t * BT, 0), (BT, BT), (1, 0))
    p_A_ak = tl.make_block_ptr(A_ak + (bos*H + i_h) * BT, (T, BT), (H*BT, 1), (i_t * BT, 0), (BT, BT), (1, 0))

    b_Aab_inv = tl.load(p_A_ab_inv, boundary_check=(0, 1))
    b_Aak = tl.load(p_A_ak, boundary_check=(0, 1))
    b_Aab_inv = tl.where(o_s[:, None] >= o_s[None, :], b_Aab_inv, 0)
    b_Aak = tl.where(o_s[:, None] > o_s[None, :], b_Aak, 0)
    # let's use tf32 here
    b_Aak = tl.dot(b_Aab_inv, b_Aak)
    # (SY 01/04) should be bf16 or tf32? To verify.
    b_Aak = b_Aak.to(v.dtype.element_ty, fp_downcast_rounding="rtne")
    b_Aab_inv = b_Aab_inv.to(ag.dtype.element_ty, fp_downcast_rounding="rtne")

    for i_k in range(tl.cdiv(K, BK)):
        p_ag = tl.make_block_ptr(ag + (bos*H + i_h) * K, (T, K), (H*K, 1), (i_t * BT, i_k * BK), (BT, BK), (1, 0))
        p_w = tl.make_block_ptr(w + (bos*H + i_h) * K, (T, K), (H*K, 1), (i_t * BT, i_k * BK), (BT, BK), (1, 0))
        b_ag = tl.load(p_ag, boundary_check=(0, 1))
        b_w = tl.dot(b_Aab_inv, b_ag)  # both bf16 or fp16
        tl.store(p_w, b_w.to(p_w.dtype.element_ty, fp_downcast_rounding="rtne"), boundary_check=(0, 1))

    for i_v in range(tl.cdiv(V, BV)):
        p_v = tl.make_block_ptr(v + (bos*H + i_h) * V, (T, V), (H*V, 1), (i_t * BT, i_v * BV), (BT, BV), (1, 0))
        p_u = tl.make_block_ptr(u + (bos*H + i_h) * V, (T, V), (H*V, 1), (i_t * BT, i_v * BV), (BT, BV), (1, 0))
        b_v = tl.load(p_v, boundary_check=(0, 1))
        b_u = tl.dot(b_Aak, b_v)  # both bf16 or fp16
        tl.store(p_u, b_u.to(p_u.dtype.element_ty, fp_downcast_rounding="rtne"), boundary_check=(0, 1))


def wu_fwd(
    ag: torch.Tensor,
    v: torch.Tensor,
    A_ak: torch.Tensor,
    A_ab_inv: torch.Tensor,
    cu_seqlens: torch.LongTensor | None,
    chunk_size: int,
) -> tuple[torch.Tensor, torch.Tensor]:
    B, T, H, K, V = *ag.shape, v.shape[-1]
    BT = chunk_size

    chunk_indices = prepare_chunk_indices(cu_seqlens, BT) if cu_seqlens is not None else None
    NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
    BK = min(max(triton.next_power_of_2(K), 16), 64)
    BV = min(max(triton.next_power_of_2(V), 16), 64)

    w = torch.empty_like(ag)
    u = torch.empty_like(v)
    wu_fwd_kernel[(NT, B * H)](
        ag=ag,
        v=v,
        A_ak=A_ak,
        A_ab_inv=A_ab_inv,
        w=w,
        u=u,
        cu_seqlens=cu_seqlens,
        chunk_indices=chunk_indices,
        T=T,
        H=H,
        K=K,
        V=V,
        BT=BT,
        BK=BK,
        BV=BV,
    )
    return w, u


def prepare_wy_repr_fwd(
    ag: torch.Tensor,
    v: torch.Tensor,
    A_ak: torch.Tensor,
    A_ab: torch.Tensor,
    cu_seqlens: torch.LongTensor | None,
    chunk_size: int = 64,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    B, T, H, _ = ag.shape
    BT = chunk_size

    chunk_indices = prepare_chunk_indices(cu_seqlens, BT) if cu_seqlens is not None else None
    NT = triton.cdiv(T, BT) if cu_seqlens is None else len(chunk_indices)
    BC = min(BT, 32)
    fwd_fn = prepare_wy_repr_fwd_kernel_chunk64 if BT == 64 else prepare_wy_repr_fwd_kernel_chunk32
    A_ab_inv = torch.empty_like(A_ab)
    fwd_fn[(NT, B * H)](
        A_ab=A_ab,
        A_ab_inv=A_ab_inv,
        cu_seqlens=cu_seqlens,
        chunk_indices=chunk_indices,
        T=T,
        H=H,
        BT=BT,
        BC=BC,
    )
    w, u = wu_fwd(
        ag=ag,
        v=v,
        A_ak=A_ak,
        A_ab_inv=A_ab_inv,
        cu_seqlens=cu_seqlens,
        chunk_size=BT,
    )
    return w, u, A_ab_inv


fwd_prepare_wy_repr = prepare_wy_repr_fwd

fwd_wu = wu_fwd



import logging
from typing import Optional, Tuple

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import device, torch_device_fn

from ..utils import MAX_GRID_SIZE_X, TOTAL_CORE_NUM

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))
device = device.name


@triton.autotune(
    configs=runtime.get_tuned_config("upsample_nearest2d"), key=["N", "C", "OH", "OW"]
)
@triton.heuristics(runtime.get_heuristic_config("upsample_nearest2d"))
@triton.jit
def upsample_nearest2d_kernel(
    ptr_o,
    ptr_i,
    N,
    C,
    OH,
    OW,
    IH,
    IW,
    reciprocal_scale_h,
    reciprocal_scale_w,
    BLOCK_SIZE: tl.constexpr,
    SAME_H: tl.constexpr,
    SAME_W: tl.constexpr,
):
    pid = tl.program_id(axis=0) + tl.program_id(axis=1) * tl.num_programs(0)
    idx = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    ow = idx % OW
    oh = idx // OW % OH
    c = idx // OW // OH % C
    n = idx // OW // OH // C % N
    if SAME_H:
        ih = oh
    else:
        # tl.floor() cannot be found in 2.3.1, using int trunc
        ih = tl.minimum((oh * reciprocal_scale_h).to(tl.int32), IH - 1)
    if SAME_W:
        iw = ow
    else:
        iw = tl.minimum((ow * reciprocal_scale_w).to(tl.int32), IW - 1)
    offset_o = ((n * C + c) * OH + oh) * OW + ow
    offset_i = ((n * C + c) * IH + ih) * IW + iw
    data = tl.load(ptr_i + offset_i)
    tl.store(ptr_o + offset_o, data)


def configs2():
    block_h = [8, 16, 32, 64, 128, 256]
    num_stage = [1, 3]
    return [
        triton.Config({"BLOCK_H": bh}, num_warps=1, num_stages=s)
        for s in num_stage
        for bh in block_h
    ]


@triton.autotune(configs=configs2(), key=["N", "C", "OH", "OW"])
@triton.jit
def upsample_nearest2d_kernel_opt(
    ptr_o,
    ptr_i,
    N,
    C,
    OH,
    OW: tl.constexpr,
    IH,
    IW: tl.constexpr,
    BLOCK_H: tl.constexpr,
):
    pid = tl.program_id(axis=0)
    num_jobs = tl.num_programs(axis=0)

    nc_nums_per_job = (N * C + num_jobs - 1) // num_jobs
    nc_begin = pid * nc_nums_per_job
    nc_end = min(nc_begin + nc_nums_per_job, N * C)

    loop_num_h = (OH + BLOCK_H - 1) // BLOCK_H
    for idx in range((nc_end - nc_begin) * loop_num_h):
        nc_idx = nc_begin + (idx // loop_num_h)
        h_idx = (idx % loop_num_h) * BLOCK_H

        init_out = nc_idx * OH * OW
        init_in = nc_idx * IH * IW

        ih = h_idx // 2 + tl.arange(0, BLOCK_H // 2)
        iw = tl.arange(0, IW)
        offset_i = init_in + ih[:, None] * IW + iw

        oh = h_idx + tl.arange(0, BLOCK_H)
        ow = tl.arange(0, OW)
        offset_o = init_out + oh[:, None] * OW + ow

        data = tl.load(ptr_i + offset_i, mask=(ih[:, None] < IH))

        tmp = (
            data.reshape(BLOCK_H // 2, OW // 2, 1)
            .broadcast_to(BLOCK_H // 2, OW // 2, 2)
            .reshape(BLOCK_H // 2, 1, OW)
        )
        tmp1 = tmp.broadcast_to(BLOCK_H // 2, 2, OW).reshape(BLOCK_H, OW)

        tl.store(ptr_o + offset_o, tmp1, mask=(oh[:, None] < OH))


@triton.autotune(configs=configs2(), key=["N", "C", "OH", "OW"])
@triton.jit
def upsample_nearest2d_kernel_opt_tile_h(
    ptr_o,
    ptr_i,
    N,
    C,
    OH,
    OW: tl.constexpr,
    IH,
    IW: tl.constexpr,
    BLOCK_H: tl.constexpr,
):
    pid = tl.program_id(axis=0)
    num_jobs = tl.num_programs(axis=0)

    start = pid * BLOCK_H
    step = BLOCK_H * num_jobs
    loop_num_h = (OH - start + step - 1) // step

    for idx in range(N * C * loop_num_h):
        nc_idx = idx // loop_num_h
        h_idx = (idx % loop_num_h) * step + start

        init_out = nc_idx * OH * OW
        init_in = nc_idx * IH * IW

        ih = h_idx // 2 + tl.arange(0, BLOCK_H // 2)
        iw = tl.arange(0, IW)
        offset_i = init_in + ih[:, None] * IW + iw

        oh = h_idx + tl.arange(0, BLOCK_H)
        ow = tl.arange(0, OW)
        offset_o = init_out + oh[:, None] * OW + ow

        data = tl.load(ptr_i + offset_i, mask=(ih[:, None] < IH))

        tmp = (
            data.reshape(BLOCK_H // 2, OW // 2, 1)
            .broadcast_to(BLOCK_H // 2, OW // 2, 2)
            .reshape(BLOCK_H // 2, 1, OW)
        )
        tmp1 = tmp.broadcast_to(BLOCK_H // 2, 2, OW).reshape(BLOCK_H, OW)

        tl.store(ptr_o + offset_o, tmp1, mask=(oh[:, None] < OH))


def upsample_nearest2d(
    input: torch.Tensor,
    output_size: Tuple[int],
    scales_h: Optional[float] = None,
    scales_w: Optional[float] = None,
) -> torch.Tensor:
    logger.debug("GEMS_CAMBRICON UPSAMPLE NEAREST2D")
    assert input.device.type == device
    assert input.ndim == 4, "The ndim of input must be 4"
    assert len(output_size) == 2, "The len of output_size must be 2"
    OH, OW = output_size
    N, C, IH, IW = input.shape
    if scales_h is not None:
        reciprocal_scale_h = 1 / scales_h
    else:
        reciprocal_scale_h = IH / OH
    if scales_w is not None:
        reciprocal_scale_w = 1 / scales_w
    else:
        reciprocal_scale_w = IW / OW
    # allocate output
    output = torch.empty((N, C, OH, OW), device=input.device, dtype=input.dtype)

    with torch_device_fn.device(input.device):
        if (
            reciprocal_scale_h == 0.5
            and reciprocal_scale_w == 0.5
            and IH / OH == 0.5
            and IW / OW == 0.5
        ):
            if N * C > 48:
                upsample_nearest2d_kernel_opt[TOTAL_CORE_NUM,](
                    output, input, N, C, OH, OW, IH, IW
                )
            else:
                upsample_nearest2d_kernel_opt_tile_h[TOTAL_CORE_NUM,](
                    output, input, N, C, OH, OW, IH, IW
                )
        else:
            total_threads = N * C * OH * OW

            # incase grid check error
            def grid_fn(META):
                num_threads = triton.cdiv(total_threads, META["BLOCK_SIZE"])
                grid_x = min(num_threads, MAX_GRID_SIZE_X)
                grid_y = triton.cdiv(num_threads, grid_x)
                return (
                    grid_x,
                    grid_y,
                )

            upsample_nearest2d_kernel[grid_fn](
                output,
                input,
                N,
                C,
                OH,
                OW,
                IH,
                IW,
                reciprocal_scale_h,
                reciprocal_scale_w,
            )

    return output



import logging
import math

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.utils import libentry, libtuner

from ..utils import TOTAL_CORE_NUM

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@libentry()
@libtuner(
    configs=[
        triton.Config(kwargs={"BLOCK_SIZE": 1024}, num_stages=3, num_warps=1),
        triton.Config(kwargs={"BLOCK_SIZE": 4096}, num_stages=3, num_warps=1),
        triton.Config(kwargs={"BLOCK_SIZE": 8192}, num_stages=3, num_warps=1),
        triton.Config(kwargs={"BLOCK_SIZE": 16384}, num_stages=3, num_warps=1),
    ],
    key=["size"],
    strategy=["log"],
)
@triton.jit
def arange_func(y_ptr, start, end, step, size, BLOCK_SIZE: tl.constexpr):
    pid = tl.program_id(axis=0)
    num_jobs = tl.num_programs(axis=0)
    block_start = pid * BLOCK_SIZE
    block_step = num_jobs * BLOCK_SIZE
    block_start = block_start
    for block_start_offset in range(block_start, size, block_step):
        offset = tl.arange(0, BLOCK_SIZE) + block_start_offset
        arange_val = offset * step + start
        tl.store(y_ptr + offset, arange_val, mask=offset < size)


def arange_start(
    start, end, step=1, *, dtype=None, layout=None, device=None, pin_memory=None
):
    logger.debug("GEMS_CAMBRICON ARANGE")
    if dtype is torch.int64:
        sgn = (step > 0) - (step < 0)
        size = (end - start + step - sgn) // step
    else:
        size = math.ceil((end - start) / step)

    assert (
        size < torch.iinfo(torch.int32).max
    ), f"Size {size} is not less than the maximum int32 value max_int32"

    grid = lambda META: (min(triton.cdiv(size, META["BLOCK_SIZE"]), TOTAL_CORE_NUM),)

    if dtype is None:
        dtype = torch.int64

    if pin_memory is None:
        pin_memory = False

    if device is None:
        device = (
            runtime.device.name
        )  # Note(Zhengzekang): Torch default value is CPU, but triton is target to GPU.

    result = torch.empty((size,), device=device, dtype=dtype, pin_memory=pin_memory)
    arange_func[grid](result, start, end, step, size)
    return result


def arange(end, *, dtype=None, layout=None, device=None, pin_memory=None):
    return arange_start(
        0, end, 1, dtype=dtype, layout=layout, device=device, pin_memory=pin_memory
    )



import logging

import triton
import triton.language as tl
from _kunlunxin.utils.codegen_config_utils import CodeGenConfig

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))

config_ = CodeGenConfig(
    512,
    (65536, 65536, 65536),
    32,
    True,
    prefer_1d_tile=True,
    buffer_size_limit=4096,
    isCloseVectorization=True,
    unroll_num=8,
)


@pointwise_dynamic(promotion_methods=[(0, "INT_TO_FLOAT")], config=config_)
@triton.jit
def sqrt_func(x):
    return tl.sqrt(x.to(tl.float32))


def sqrt(A):
    logger.debug("GEMS SQRT")
    return sqrt_func(A)


def sqrt_(A):
    logger.debug("GEMS SQRT_")
    sqrt_func(A, out0=A)
    return A



"""
Component Placement Execution Strategy

Whole components distributed across multiple GPUs.
Each component executes on its assigned GPU with tensor transfers between.

Two variants:
- ComponentPlacementStrategy (component_placement): Eager — weights stay in VRAM permanently.
- ComponentPlacementLazyStrategy (component_placement_lazy): Lazy — weights swap in/out
  between execution phases. Used when total model exceeds combined VRAM but each
  phase (pre_loop, loop, post_loop) fits.

Key features:
- Cross-device tensor transfers (automatic)
- Async transfers when available
- State tracking for efficient transfers
"""

from typing import Dict, Any, Optional, Set
import torch
from neurobrix.core.device_utils import device_empty_cache

from .base import ExecutionStrategy, StrategyContext


class ComponentPlacementStrategy(ExecutionStrategy):
    """
    Component placement across multiple GPUs.

    Each component (text_encoder, transformer, vae) is on a different GPU.
    Tensors are automatically transferred between devices at component boundaries.
    """

    def __init__(self, context: StrategyContext, strategy_name: str = "component_placement"):
        super().__init__(context, strategy_name)

        self.devices = context.get_all_devices()
        self.use_async = True  # Always use async when available

        # Map components to their devices
        self.component_devices: Dict[str, str] = {}
        for comp_name, alloc_info in context.allocations.items():
            if isinstance(alloc_info, dict):
                device = alloc_info.get('device')
                assert device is not None, f"Allocation dict missing 'device' key for {comp_name}"
            else:
                assert isinstance(alloc_info, tuple), f"Allocation must be dict or tuple, got {type(alloc_info)}"
                device = alloc_info[0]
            self.component_devices[comp_name] = device

        self.state_device: Optional[str] = None

    def execute_component(
        self,
        component_name: str,
        phase: str,
        inputs: Optional[Dict[str, Any]] = None,
    ) -> Optional[Dict[str, Any]]:
        """Execute component on its assigned GPU."""
        executor = self.context.component_executors.get(component_name)
        if executor is None:
            raise RuntimeError(f"ZERO FALLBACK: No executor for '{component_name}'")

        target_device = self.component_devices.get(component_name)
        if target_device is None:
            raise RuntimeError(f"ZERO FALLBACK: No device assignment for '{component_name}'")

        self.load_weights(component_name)

        if inputs:
            inputs = self.prepare_inputs(component_name, inputs)

        return executor.run(inputs or {})

    def prepare_inputs(self, component_name: str, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Transfer inputs to component's device."""
        target_device = self.component_devices.get(component_name)
        if target_device is None:
            return inputs

        transferred = self.transfer_dict(inputs, target_device, self.use_async)
        if self.use_async:
            self.synchronize_device(target_device)
        return transferred

    def handle_outputs(
        self,
        component_name: str,
        outputs: Dict[str, Any],
        target_device: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Handle outputs - transfer to target device if specified."""
        source_device = self.component_devices.get(component_name)
        if target_device and target_device != source_device:
            outputs = self.transfer_dict(outputs, target_device, self.use_async)
            if self.use_async:
                self.synchronize_device(target_device)
        return outputs

    def transfer_model_output(self, model_output: torch.Tensor, state_tensor: torch.Tensor) -> torch.Tensor:
        """Transfer model output to same device as state tensor."""
        if model_output.device == state_tensor.device:
            return model_output
        target_device = str(state_tensor.device)
        transferred = self.transfer_tensor(model_output, target_device, self.use_async)
        if self.use_async:
            self.synchronize_device(target_device)
        return transferred

    def get_component_device(self, component_name: str) -> str:
        """Get device for component."""
        device = self.component_devices.get(component_name)
        if device is None:
            raise RuntimeError(f"ZERO FALLBACK: No device for component '{component_name}'")
        return device

    def set_state_device(self, device: str) -> None:
        self.state_device = device

    def get_state_device(self) -> Optional[str]:
        return self.state_device

    def ensure_on_device(self, tensor: torch.Tensor, device: str) -> torch.Tensor:
        return self.transfer_tensor(tensor, device, self.use_async)


class ComponentPlacementLazyStrategy(ComponentPlacementStrategy):
    """
    Component placement with lazy weight swapping between phases.

    Weights are loaded before each execution phase and unloaded after.
    Allows models where total weights exceed combined VRAM, as long as
    each phase (pre_loop, loop, post_loop) fits.
    """

    def __init__(self, context: StrategyContext, strategy_name: str = "component_placement_lazy"):
        super().__init__(context, strategy_name)
        self._phase_loaded: Set[str] = set()

    def execute_component(
        self,
        component_name: str,
        phase: str,
        inputs: Optional[Dict[str, Any]] = None,
    ) -> Optional[Dict[str, Any]]:
        """Execute component with lazy load/unload around execution."""
        if component_name not in self._phase_loaded:
            self.load_weights(component_name)
            self._phase_loaded.add(component_name)
        return super().execute_component(component_name, phase, inputs)

    def end_phase(self, phase: str) -> None:
        """Signal end of execution phase — unload weights loaded during this phase."""
        for comp_name in list(self._phase_loaded):
            self.unload_weights(comp_name)
        self._phase_loaded.clear()
        device_empty_cache(device)



import logging

import torch
import triton
import triton.language as tl

from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry, tl_extra_shim

from ..utils import TOTAL_CORE_NUM

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))
rsqrt = tl_extra_shim.rsqrt


def group_norm_kernel_opt_prune(configs, named_args, **kwargs):
    pruned_configs = []
    hw = kwargs["HW"]
    num_groups = named_args["num_groups"]
    all_sizes = []
    for config in configs:
        BLOCK_HW_SIZE = config.kwargs["BLOCK_HW_SIZE"]
        if BLOCK_HW_SIZE not in all_sizes:
            all_sizes.append(BLOCK_HW_SIZE)

    for config in configs:
        BLOCK_HW_SIZE = config.kwargs["BLOCK_HW_SIZE"]
        SPLIT = config.kwargs["SPLIT"]
        if (hw > 4096) and (BLOCK_HW_SIZE >= 4096) and (SPLIT <= 1):
            pruned_configs.append(config)
        elif (BLOCK_HW_SIZE >= hw) and (SPLIT <= num_groups):
            not_step_bigger = False
            for size in all_sizes:
                if (size < BLOCK_HW_SIZE) and (size > hw):
                    not_step_bigger = True
            if not not_step_bigger:
                pruned_configs.append(config)
    return pruned_configs


@libentry()
@triton.autotune(
    configs=[
        triton.Config({"SPLIT": s, "BLOCK_HW_SIZE": size}, num_stages=3, num_warps=1)
        for size in [64, 256, 512, 1024, 2048, 4096, 5120]
        for s in [1, 4, 6, 8, 16]
    ],
    key=["X", "group_size", "C", "HW", "num_groups"],
    prune_configs_by={"early_config_prune": group_norm_kernel_opt_prune},
)
@triton.jit(do_not_specialize=["eps"])
def group_norm_kernel_opt(
    X,
    Y,
    W,
    B,
    Mean,
    Rstd,
    group_size,
    C,
    num_groups,
    eps,
    HW: tl.constexpr,
    BLOCK_GROUP_SIZE: tl.constexpr,
    BLOCK_HW_SIZE: tl.constexpr,
    SPLIT: tl.constexpr,
):
    pid = tl.program_id(0)
    div_v = tl.cdiv(num_groups, SPLIT)
    div_mod = num_groups % SPLIT
    split_group = pid % div_v
    split_n = pid // div_v
    real_num_elements = group_size * HW

    group_offset = tl.arange(0, BLOCK_GROUP_SIZE)
    hw_offset = tl.arange(0, BLOCK_HW_SIZE)
    if BLOCK_HW_SIZE >= HW:
        hw_offset = tl.arange(0, HW)
    hw_iter = tl.cdiv(HW, BLOCK_HW_SIZE)

    if W is None:
        W_ptr = None
    else:
        W_ptr = W + split_group * SPLIT * group_size
    if B is None:
        B_ptr = None
    else:
        B_ptr = B + split_group * SPLIT * group_size

    Mean_ptr = Mean + split_n * num_groups + split_group * SPLIT
    Rstd_ptr = Rstd + split_n * num_groups + split_group * SPLIT

    xy_offset = (
        split_n * C * HW
        + split_group * SPLIT * real_num_elements
        + group_offset[:, None] * HW
        + hw_offset[None, :]
    )

    ub = SPLIT
    if (div_mod != 0) and ((split_group + 1) == div_v):
        ub = div_mod
    for idx in range(0, ub):
        if BLOCK_HW_SIZE >= HW:
            tmp = tl.load(X + xy_offset, cache_modifier=".cg").to(tl.float32)
            mean = tl.sum(tmp) / real_num_elements
            x = tmp - mean
            var = tl.sum(x * x) / real_num_elements
            var = tl.rsqrt(var + eps)

            tl.store(Mean_ptr + idx, mean)
            tl.store(Rstd_ptr + idx, var)

            if W_ptr is None:
                weight = 1
            else:
                weight = tl.load(W_ptr + group_offset, cache_modifier=".cg")[:, None]
            if B_ptr is None:
                bias = 0
            else:
                bias = tl.load(B_ptr + group_offset, cache_modifier=".cg")[:, None]
            tmp = (tmp - mean) * var
            tmp = tmp * weight + bias
            tl.store(Y + xy_offset, tmp)
        else:
            mean = tl.zeros([BLOCK_GROUP_SIZE, BLOCK_HW_SIZE], tl.float32)
            var = tl.zeros([BLOCK_GROUP_SIZE, BLOCK_HW_SIZE], tl.float32)
            for idy in range(0, hw_iter):
                xy_mask = (
                    group_offset[:, None] < group_size
                    and (idy * BLOCK_HW_SIZE + hw_offset[None, :]) < HW
                )
                tmp = tl.load(
                    X + idy * BLOCK_HW_SIZE + xy_offset,
                    mask=xy_mask,
                    other=0.0,
                    cache_modifier=".cg",
                ).to(tl.float32)
                mean += tmp
                var += tmp * tmp
            mean = tl.sum(mean) / real_num_elements
            var = tl.sum(var) / real_num_elements - (mean * mean)
            var = tl.rsqrt(var + eps)
            tl.store(Mean_ptr + idx, mean)
            tl.store(Rstd_ptr + idx, var)

            if W_ptr is None:
                weight = 1
            else:
                weight = tl.load(W_ptr + group_offset, cache_modifier=".cg")[:, None]
            if B_ptr is None:
                bias = 0
            else:
                bias = tl.load(B_ptr + group_offset, cache_modifier=".cg")[:, None]

            for idy in range(0, hw_iter):
                xy_mask = (
                    group_offset[:, None] < group_size
                    and (idy * BLOCK_HW_SIZE + hw_offset[None, :]) < HW
                )
                tmp = tl.load(
                    X + idy * BLOCK_HW_SIZE + xy_offset,
                    mask=xy_mask,
                    other=0.0,
                    cache_modifier=".cg",
                ).to(tl.float32)
                tmp = (tmp - mean) * var
                tmp = tmp * weight + bias
                tl.store(Y + idy * BLOCK_HW_SIZE + xy_offset, tmp, mask=xy_mask)

        xy_offset += real_num_elements
        group_offset += group_size


def group_norm_backward_kernel_opt_prune(configs, named_args, **kwargs):
    pruned_configs = []
    hw = kwargs["HW"]
    all_sizes = []
    for config in configs:
        BLOCK_HW_SIZE = config.kwargs["BLOCK_HW_SIZE"]
        if BLOCK_HW_SIZE not in all_sizes:
            all_sizes.append(BLOCK_HW_SIZE)
    for config in configs:
        BLOCK_HW_SIZE = config.kwargs["BLOCK_HW_SIZE"]
        SPLIT = config.kwargs["SPLIT"]
        if (hw > 2048) and (BLOCK_HW_SIZE >= 2048) and (SPLIT <= 1):
            pruned_configs.append(config)
        elif BLOCK_HW_SIZE > hw:
            not_step_bigger = False
            for size in all_sizes:
                if (size < BLOCK_HW_SIZE) and (size > hw):
                    not_step_bigger = True
            if not not_step_bigger:
                pruned_configs.append(config)
    return pruned_configs


@libentry()
@triton.autotune(
    configs=[
        triton.Config({"SPLIT": s, "BLOCK_HW_SIZE": size}, num_stages=3, num_warps=1)
        for s in [1, 4, 6, 8]
        for size in [64, 256, 512, 1024, 2048]
    ],
    prune_configs_by={"early_config_prune": group_norm_backward_kernel_opt_prune},
    key=["X", "group_size", "C", "HW", "num_groups"],
)
@triton.jit()
def group_norm_backward_kernel_opt(
    grad_y,
    X,
    W,
    Mean,
    Rstd,
    num_groups,
    group_size,
    grad_x,
    C,
    HW: tl.constexpr,
    BLOCK_GROUP_SIZE: tl.constexpr,
    BLOCK_HW_SIZE: tl.constexpr,
    SPLIT: tl.constexpr,
):
    pid = tl.program_id(0)
    div_v = tl.cdiv(num_groups, SPLIT)
    div_mod = num_groups % SPLIT
    split_group = pid % div_v
    split_n = pid // div_v
    real_num_elements = group_size * HW
    hw_iter = tl.cdiv(HW, BLOCK_HW_SIZE)

    group_offset = tl.arange(0, BLOCK_GROUP_SIZE)
    if BLOCK_HW_SIZE >= HW:
        hw_offset = tl.arange(0, HW)
    else:
        hw_offset = tl.arange(0, BLOCK_HW_SIZE)

    if W is None:
        W_ptr = None
    else:
        W_ptr = W + split_group * SPLIT * group_size

    Mean_ptr = Mean + split_n * num_groups + split_group * SPLIT
    Rstd_ptr = Rstd + split_n * num_groups + split_group * SPLIT

    xy_offset = (
        split_n * real_num_elements * num_groups
        + split_group * SPLIT * real_num_elements
        + group_offset[:, None] * HW
        + hw_offset[None, :]
    )

    ub = SPLIT
    if (div_mod != 0) and ((split_group + 1) == div_v):
        ub = div_mod
    for idx in range(0, ub):
        wb_mask = group_offset < C

        if W_ptr is None:
            weight = 1
        else:
            weight = tl.load(
                W_ptr + group_offset, mask=wb_mask, other=0.0, cache_modifier=".cg"
            ).to(tl.float32)[:, None]
        rstd = tl.load(Rstd_ptr + idx).to(tl.float32)
        mean = tl.load(Mean_ptr + idx).to(tl.float32)

        if BLOCK_HW_SIZE >= HW:
            dY_val = tl.load(grad_y + xy_offset, cache_modifier=".cg").to(tl.float32)
            X_val = tl.load(X + xy_offset, cache_modifier=".cg").to(tl.float32)

            x_hat = rstd * (X_val - mean)
            dx_hat = weight * dY_val

            grad_dx_hat_sum = tl.sum(dx_hat)
            grad_x_hat_sum = tl.sum(dx_hat * x_hat)

            dx = rstd * (
                dx_hat - (grad_dx_hat_sum + x_hat * grad_x_hat_sum) / real_num_elements
            )

            tl.store(grad_x + xy_offset, dx)
        else:
            grad_dx_hat_accum = tl.zeros([BLOCK_GROUP_SIZE, BLOCK_HW_SIZE], tl.float32)
            grad_x_hat_accum = tl.zeros([BLOCK_GROUP_SIZE, BLOCK_HW_SIZE], tl.float32)

            for idy in range(0, hw_iter):
                xy_mask = (group_offset[:, None] < C) & (
                    (idy * BLOCK_HW_SIZE + hw_offset[None, :]) < HW
                )
                dY_val = tl.load(
                    grad_y + idy * BLOCK_HW_SIZE + xy_offset,
                    mask=xy_mask,
                    other=0.0,
                    cache_modifier=".cg",
                ).to(tl.float32)
                X_val = tl.load(
                    X + idy * BLOCK_HW_SIZE + xy_offset,
                    mask=xy_mask,
                    other=0.0,
                    cache_modifier=".cg",
                ).to(tl.float32)

                x_hat = tl.where(xy_mask, rstd * (X_val - mean), 0.0)
                dx_hat = weight * dY_val
                grad_dx_hat_accum += dx_hat
                grad_x_hat_accum += dx_hat * x_hat

            grad_dx_hat_total = tl.sum(grad_dx_hat_accum)
            grad_x_hat_total = tl.sum(grad_x_hat_accum)

            for idy in range(0, hw_iter):
                xy_mask = (group_offset[:, None] < C) & (
                    (idy * BLOCK_HW_SIZE + hw_offset[None, :]) < HW
                )
                dY_val = tl.load(
                    grad_y + idy * BLOCK_HW_SIZE + xy_offset,
                    mask=xy_mask,
                    other=0.0,
                    cache_modifier=".cg",
                ).to(tl.float32)
                X_val = tl.load(
                    X + idy * BLOCK_HW_SIZE + xy_offset,
                    mask=xy_mask,
                    other=0.0,
                    cache_modifier=".cg",
                ).to(tl.float32)

                x_hat = tl.where(xy_mask, rstd * (X_val - mean), 0.0)
                dx_hat = weight * dY_val
                dx = rstd * (
                    dx_hat
                    - (grad_dx_hat_total + x_hat * grad_x_hat_total) / real_num_elements
                )

                tl.store(grad_x + idy * BLOCK_HW_SIZE + xy_offset, dx, mask=xy_mask)

        xy_offset += real_num_elements
        group_offset += group_size


def weight_bias_backward_kernel_opt_prune(configs, named_args, **kwargs):
    pruned_configs = []
    pruned_configs_cached = []
    n = named_args["N"]
    hw = kwargs["HW"]
    all_sizes = []
    for config in configs:
        BLOCK_HW_SIZE = config.kwargs["BLOCK_HW_SIZE"]
        if BLOCK_HW_SIZE not in all_sizes:
            all_sizes.append(BLOCK_HW_SIZE)
    for config in configs:
        BLOCK_HW_SIZE = config.kwargs["BLOCK_HW_SIZE"]
        BLOCK_N = config.kwargs["BLOCK_N"]
        if (hw > 2048) and (BLOCK_HW_SIZE >= 2048) and (BLOCK_N <= 4):
            pruned_configs_cached.append(config)
        elif BLOCK_HW_SIZE > hw:
            not_step_bigger = False
            for size in all_sizes:
                if (size < BLOCK_HW_SIZE) and (size > hw):
                    not_step_bigger = True
            if not not_step_bigger:
                pruned_configs_cached.append(config)
    # remove some block n
    for config in pruned_configs_cached:
        block_n = config.kwargs["BLOCK_N"]
        if n % block_n == 0:
            pruned_configs.append(config)
    return pruned_configs


@libentry()
@triton.autotune(
    configs=[
        triton.Config({"BLOCK_N": bn, "BLOCK_HW_SIZE": size}, num_stages=3, num_warps=1)
        for bn in [1, 4, 8, 16]
        for size in [512, 1024, 2048]
    ],
    prune_configs_by={"early_config_prune": weight_bias_backward_kernel_opt_prune},
    key=["X", "N", "C", "HW", "num_groups"],
)
@triton.jit
def weight_bias_backward_kernel_opt(
    dY,
    X,
    Mean,
    Rstd,
    dW,
    dB,
    num_groups,
    group_size,
    N,
    C,
    HW: tl.constexpr,
    BLOCK_N: tl.constexpr,
    BLOCK_HW_SIZE: tl.constexpr,
):
    pid = tl.program_id(0)
    pnum = tl.num_programs(axis=0)
    C_SPLIT = tl.cdiv(C, pnum)
    N_SPLIT = tl.cdiv(N, BLOCK_N)
    hw_iter = tl.cdiv(HW, BLOCK_HW_SIZE)

    n_offset = tl.arange(0, BLOCK_N)
    hw_offset = tl.arange(0, BLOCK_HW_SIZE)
    if BLOCK_HW_SIZE >= HW:
        hw_offset = tl.arange(0, HW)

    lb = pid * C_SPLIT
    ub = tl.minimum((pid + 1) * C_SPLIT, C)
    for c_start in range(lb, ub):
        if BLOCK_HW_SIZE >= HW:
            dY_ptr = dY + c_start * HW + n_offset[:, None] * C * HW + hw_offset[None, :]
            x_ptr = X + c_start * HW + n_offset[:, None] * C * HW + hw_offset[None, :]
            grad_y = tl.load(dY_ptr, cache_modifier=".cg").to(tl.float32)

            x = tl.load(x_ptr, cache_modifier=".cg")
            x_f32 = x.to(tl.float32)

            mean_ptr = Mean + c_start // group_size + n_offset * num_groups
            rstd_ptr = Rstd + c_start // group_size + n_offset * num_groups

            mean = tl.load(mean_ptr, cache_modifier=".cg").to(tl.float32)[:, None]
            rstd = tl.load(rstd_ptr, cache_modifier=".cg").to(tl.float32)[:, None]

            dB_val = tl.sum(grad_y)
            dW_val = tl.sum((x_f32 - mean) * rstd * grad_y)

            for n_start in range(1, N_SPLIT):
                new_n_offset = n_start * BLOCK_N + n_offset

                dY_ptr = (
                    dY
                    + c_start * HW
                    + new_n_offset[:, None] * C * HW
                    + hw_offset[None, :]
                )
                x_ptr = (
                    X
                    + c_start * HW
                    + new_n_offset[:, None] * C * HW
                    + hw_offset[None, :]
                )
                grad_y = tl.load(dY_ptr, cache_modifier=".cg").to(tl.float32)

                x = tl.load(x_ptr, cache_modifier=".cg")
                x_f32 = x.to(tl.float32)

                mean_ptr = Mean + c_start // group_size + new_n_offset * num_groups
                rstd_ptr = Rstd + c_start // group_size + new_n_offset * num_groups

                mean = tl.load(mean_ptr, cache_modifier=".cg").to(tl.float32)[:, None]
                rstd = tl.load(rstd_ptr, cache_modifier=".cg").to(tl.float32)[:, None]

                dB_val += tl.sum(grad_y)
                dW_val += tl.sum((x_f32 - mean) * rstd * grad_y)

            if dW is not None:
                tl.store(dW + c_start, dW_val)
            if dB is not None:
                tl.store(dB + c_start, dB_val)
        else:
            xy_mask = (n_offset[:, None] < N) & (hw_offset[None, :] < HW)

            dY_ptr = dY + c_start * HW + n_offset[:, None] * C * HW + hw_offset[None, :]
            x_ptr = X + c_start * HW + n_offset[:, None] * C * HW + hw_offset[None, :]
            grad_y = tl.load(dY_ptr, cache_modifier=".cg").to(tl.float32)

            x = tl.load(x_ptr, cache_modifier=".cg")
            x_f32 = x.to(tl.float32)

            mean_ptr = Mean + c_start // group_size + n_offset * num_groups
            rstd_ptr = Rstd + c_start // group_size + n_offset * num_groups

            mean = tl.load(mean_ptr, cache_modifier=".cg").to(tl.float32)[:, None]
            rstd = tl.load(rstd_ptr, cache_modifier=".cg").to(tl.float32)[:, None]

            dB_val = tl.sum(grad_y)
            dW_val = tl.sum((x_f32 - mean) * rstd * grad_y)

            for idx in range(1, hw_iter):
                xy_mask = (n_offset[:, None] < N) & (
                    (idx * BLOCK_HW_SIZE + hw_offset[None, :]) < HW
                )
                dY_ptr = (
                    dY
                    + c_start * HW
                    + n_offset[:, None] * C * HW
                    + hw_offset[None, :]
                    + idx * BLOCK_HW_SIZE
                )
                x_ptr = (
                    X
                    + c_start * HW
                    + n_offset[:, None] * C * HW
                    + hw_offset[None, :]
                    + idx * BLOCK_HW_SIZE
                )

                grad_y = tl.load(
                    dY_ptr, mask=xy_mask, other=0.0, cache_modifier=".cg"
                ).to(tl.float32)
                x = tl.load(x_ptr, mask=xy_mask, other=0.0, cache_modifier=".cg")
                x_f32 = x.to(tl.float32)
                dB_val += tl.sum(grad_y)
                x_f32 = tl.where(xy_mask, x_f32 - mean, 0.0)
                dW_val += tl.sum(x_f32 * rstd * grad_y)

            for n_start in range(1, N_SPLIT):
                new_n_offset = n_start * BLOCK_N + n_offset
                xy_mask = (new_n_offset[:, None] < N) & (hw_offset[None, :] < HW)

                dY_ptr = (
                    dY
                    + c_start * HW
                    + new_n_offset[:, None] * C * HW
                    + hw_offset[None, :]
                )
                x_ptr = (
                    X
                    + c_start * HW
                    + new_n_offset[:, None] * C * HW
                    + hw_offset[None, :]
                )
                grad_y = tl.load(
                    dY_ptr, mask=xy_mask, other=0.0, cache_modifier=".cg"
                ).to(tl.float32)

                x = tl.load(x_ptr, mask=xy_mask, other=0.0, cache_modifier=".cg")
                x_f32 = x.to(tl.float32)

                mean_ptr = Mean + c_start // group_size + new_n_offset * num_groups
                rstd_ptr = Rstd + c_start // group_size + new_n_offset * num_groups

                mean = tl.load(mean_ptr, cache_modifier=".cg").to(tl.float32)[:, None]
                rstd = tl.load(rstd_ptr, cache_modifier=".cg").to(tl.float32)[:, None]

                dB_val += tl.sum(grad_y)
                dW_val += tl.sum((x_f32 - mean) * rstd * grad_y)

                for idx in range(1, hw_iter):
                    xy_mask = (new_n_offset[:, None] < N) & (
                        (idx * BLOCK_HW_SIZE + hw_offset[None, :]) < HW
                    )
                    dY_ptr = (
                        dY
                        + c_start * HW
                        + new_n_offset[:, None] * C * HW
                        + hw_offset[None, :]
                        + idx * BLOCK_HW_SIZE
                    )
                    x_ptr = (
                        X
                        + c_start * HW
                        + new_n_offset[:, None] * C * HW
                        + hw_offset[None, :]
                        + idx * BLOCK_HW_SIZE
                    )

                    grad_y = tl.load(
                        dY_ptr, mask=xy_mask, other=0.0, cache_modifier=".cg"
                    ).to(tl.float32)
                    x = tl.load(x_ptr, mask=xy_mask, other=0.0, cache_modifier=".cg")
                    x_f32 = x.to(tl.float32)
                    dB_val += tl.sum(grad_y)
                    x_f32 = tl.where(xy_mask, x_f32 - mean, 0.0)
                    dW_val += tl.sum(x_f32 * rstd * grad_y)
            if dW is not None:
                tl.store(dW + c_start, dW_val)
            if dB is not None:
                tl.store(dB + c_start, dB_val)


def group_norm(input, weight, bias, N, C, HxW, group, eps=1e-05):
    logger.debug("GEMS_CAMBRICON GROUPNORM FORWARD")
    group_size = C // group
    input = input.contiguous()
    if weight is not None:
        weight = weight.contiguous()
    if bias is not None:
        bias = bias.contiguous()
    y = torch.empty_like(input)
    mean = torch.empty((N, group), dtype=input.dtype, device=input.device)
    rstd = torch.empty((N, group), dtype=input.dtype, device=input.device)
    grid = lambda meta: (N * triton.cdiv(group, meta["SPLIT"]),)

    with torch_device_fn.device(input.device):
        group_norm_kernel_opt[grid](
            input,
            y,
            weight,
            bias,
            mean,
            rstd,
            group_size,
            C,
            group,
            eps,
            HW=HxW,
            BLOCK_GROUP_SIZE=group_size,
        )
    return y, mean, rstd


def group_norm_backward(
    grad_out, input, mean, rstd, weight, N, C, HxW, group, output_mask
):
    logger.debug("GEMS_CAMBRICON GROUPNORM BACKWARD")

    grad_out = grad_out.contiguous()
    input = input.contiguous()
    mean = mean.contiguous()
    rstd = rstd.contiguous()
    weight = None if weight is None else weight.contiguous()
    group_size = triton.cdiv(C, group)

    if output_mask[0]:
        grad_inp = torch.empty_like(input)
        grid = lambda meta: (N * triton.cdiv(group, meta["SPLIT"]),)
        with torch_device_fn.device(input.device):
            group_norm_backward_kernel_opt[grid](
                grad_out,
                input,
                weight,
                mean,
                rstd,
                group,
                group_size,
                grad_inp,
                C,
                HW=HxW,
                BLOCK_GROUP_SIZE=group_size,
            )
    else:
        grad_inp = None

    if output_mask[1] is False and output_mask[2] is False:
        return grad_inp, None, None

    weight_grad = torch.empty_like(weight) if output_mask[1] else None
    bias_grad = torch.empty_like(weight) if output_mask[2] else None
    with torch_device_fn.device(input.device):
        weight_bias_backward_kernel_opt[(TOTAL_CORE_NUM, 1, 1)](
            grad_out,
            input,
            mean,
            rstd,
            weight_grad,
            bias_grad,
            group,
            group_size,
            N,
            C,
            HW=HxW,
        )
    return grad_inp, weight_grad, bias_grad



import logging

import torch

from flag_gems.ops.copy import copy
from flag_gems.utils.shape_utils import has_internal_overlapping

logger = logging.getLogger(f'flag_gems.runtime._ascend.ops.{__name__.split(".")[-1]}')


def select_scatter(inp, src, dim, index):
    logger.debug("GEMS_ASCEND SELECT_SCATTER")
    assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
    assert index >= -inp.size(dim) and index < inp.size(dim), "Invalid index"
    dim = dim % inp.ndim
    index = index % inp.size(dim)

    valid_shape = list(inp.shape)
    del valid_shape[dim]
    assert (
        list(src.shape) == valid_shape
    ), "Expected src to have a size equal to the slice of self"

    if has_internal_overlapping(inp):
        out = torch.empty(inp.size(), dtype=inp.dtype, device=inp.device)
    else:
        out = torch.empty_strided(
            inp.size(), inp.stride(), dtype=inp.dtype, device=inp.device
        )

    copy(inp, out0=out)
    indices = [slice(None)] * inp.ndim
    indices[dim] = index
    copy(src, out0=out[indices])

    return out



"""
Universal MatMul Kernel - NeuroBrix Standard (Triton Optimized)
Source: FlagGems / Liger Logic
"""
import torch
import triton
import triton.language as tl
from neurobrix.kernels.registry import register_kernel

# --- TRITON KERNEL (PURE 2D) ---
@triton.jit
def _matmul_kernel(
    a_ptr, b_ptr, c_ptr,
    M, N, K,
    stride_am, stride_ak,
    stride_bk, stride_bn,
    stride_cm, stride_cn,
    BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, BLOCK_SIZE_K: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
):
    pid = tl.program_id(axis=0)
    num_pid_m = tl.cdiv(M, BLOCK_SIZE_M)
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)
    num_pid_in_group = GROUP_SIZE_M * num_pid_n
    group_id = pid // num_pid_in_group
    first_pid_m = group_id * GROUP_SIZE_M
    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)
    pid_m = first_pid_m + (pid % group_size_m)
    pid_n = (pid % num_pid_in_group) // group_size_m

    offs_am = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M
    offs_bn = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)) % N
    offs_k = tl.arange(0, BLOCK_SIZE_K)
    
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)
    b_ptrs = b_ptr + (offs_k[:, None] * stride_bk + offs_bn[None, :] * stride_bn)

    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K)):
        a = tl.load(a_ptrs, mask=offs_k[None, :] < K - k * BLOCK_SIZE_K, other=0.0)
        b = tl.load(b_ptrs, mask=offs_k[:, None] < K - k * BLOCK_SIZE_K, other=0.0)
        accumulator += tl.dot(a, b)
        a_ptrs += BLOCK_SIZE_K * stride_ak
        b_ptrs += BLOCK_SIZE_K * stride_bk

    offs_cm = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    c_ptrs = c_ptr + stride_cm * offs_cm[:, None] + stride_cn * offs_cn[None, :]
    c_mask = (offs_cm[:, None] < M) & (offs_cn[None, :] < N)
    tl.store(c_ptrs, accumulator, mask=c_mask)

# --- OP LAUNCH (UNIVERSAL FLATTENING) ---

def _matmul_2d(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    """2D matmul using Triton kernel."""
    M, K = a.shape
    K2, N = b.shape
    assert K == K2, f"Shape mismatch: {a.shape} @ {b.shape}"

    c = torch.empty((M, N), device=a.device, dtype=a.dtype)

    BLOCK_SIZE_M = 64
    BLOCK_SIZE_N = 64
    BLOCK_SIZE_K = 32
    GROUP_SIZE_M = 8

    grid = (triton.cdiv(M, BLOCK_SIZE_M) * triton.cdiv(N, BLOCK_SIZE_N),)

    with torch.cuda.device(a.device):
        _matmul_kernel[grid](
            a, b, c,
            M, N, K,
            a.stride(0), a.stride(1),
            b.stride(0), b.stride(1),
            c.stride(0), c.stride(1),
            BLOCK_SIZE_M=BLOCK_SIZE_M,
            BLOCK_SIZE_N=BLOCK_SIZE_N,
            BLOCK_SIZE_K=BLOCK_SIZE_K,
            GROUP_SIZE_M=GROUP_SIZE_M,
            num_warps=4,
        )

    return c


@register_kernel(family="audio", vendor="nvidia", tier="common", op_name="matmul")
def matmul_kernel(a: torch.Tensor, b: torch.Tensor,
                  trans_a: bool = False, trans_b: bool = False) -> torch.Tensor:
    """Matrix multiply - pure Triton."""
    assert a.is_cuda and b.is_cuda

    if trans_a:
        a = a.transpose(-2, -1)
    if trans_b:
        b = b.transpose(-2, -1)

    a = a.contiguous()
    b = b.contiguous()

    # Handle batched matmul by flattening batch dims
    if a.ndim > 2 or b.ndim > 2:
        # Broadcast and flatten to 3D then iterate
        a_shape = a.shape
        b_shape = b.shape

        # Flatten batch dimensions
        if a.ndim == 2:
            a = a.unsqueeze(0)
        if b.ndim == 2:
            b = b.unsqueeze(0)

        # Flatten all batch dims
        a_batch = a.shape[:-2]
        b_batch = b.shape[:-2]

        a_flat = a.reshape(-1, a.shape[-2], a.shape[-1])
        b_flat = b.reshape(-1, b.shape[-2], b.shape[-1])

        # Broadcast batch dims
        if a_flat.shape[0] == 1:
            a_flat = a_flat.expand(b_flat.shape[0], -1, -1).contiguous()
        if b_flat.shape[0] == 1:
            b_flat = b_flat.expand(a_flat.shape[0], -1, -1).contiguous()

        batch_size = a_flat.shape[0]
        M, K = a_flat.shape[-2], a_flat.shape[-1]
        N = b_flat.shape[-1]

        c_flat = torch.empty((batch_size, M, N), device=a.device, dtype=a.dtype)

        # Run 2D matmul for each batch
        for i in range(batch_size):
            c_flat[i] = _matmul_2d(a_flat[i], b_flat[i])

        # Reshape output
        out_batch = list(torch.broadcast_shapes(a_batch, b_batch))
        return c_flat.view(out_batch + [M, N])

    # Simple 2D case
    return _matmul_2d(a, b)




import torch
import triton
import triton.language as tl

@triton.jit
def recompute_mask_kernel(mask, B, H, M, N, dropout_p, seed, offset):
    row, b, h = tl.program_id(0), tl.program_id(1), tl.program_id(2)
    offs_base = b * H * M * N + h * M * N + row * N
    BLOCK: tl.constexpr = 1024
    offs_base += tl.arange(0, BLOCK)
    for start_n in range(0, N, BLOCK):
        offs = start_n + offs_base
        rng_offs = offset + offs
        pmask = tl.rand(seed, rng_offs, n_rounds=6) > dropout_p
        row_mask = start_n + tl.arange(0, BLOCK) < N
        tl.store(mask + offs, pmask, mask=row_mask)

def recompute_mask(B, H, M, N, dropout_p, seed, offset, device):
    mask = torch.full((B, H, M, N), True, dtype=torch.bool, device=device)
    if dropout_p == 0:
        return mask
    grid = (M, B, H)
    with torch.cuda.device(device):
        recompute_mask_kernel[grid](mask, B, H, M, N, dropout_p, seed, offset)
    return mask



"""
Symbolic Shape Resolver for NeuroBrix Runtime.

Resolves symbolic shapes to concrete values at runtime based on actual input tensors.

Supports SymInt expression trees for automatic propagation.
Handles both SymInt dicts and string expressions for backward compatibility.

ZERO HARDCODE: Shape values come from actual inputs, not config.
ZERO FALLBACK: Missing symbols raise explicit errors.

Usage:
    resolver = SymbolicShapeResolver(dag["symbolic_context"])
    resolver.bind_from_inputs(inputs, dag["tensors"])

    # Resolve a shape (string or SymInt)
    concrete_shape = resolver.resolve(["s0", 4, 128, 128])
    # Returns: [2, 4, 128, 128] if s0=2
"""

import torch
import logging
import re
from typing import Dict, List, Any, Optional, Union, Tuple

# Import SymInt support (from core/runtime, no trace/ dependency)
try:
    from neurobrix.core.runtime.symint import SymInt
except ImportError:
    SymInt = None  # Fallback for when running standalone

logger = logging.getLogger(__name__)


class ShapeResolutionError(Exception):
    """Raised when symbolic shape resolution fails."""
    pass


class SymbolicShapeResolver:
    """
    Resolves symbolic shapes at runtime.

    Workflow:
    1. Initialize with symbolic_context from graph.json
    2. Bind symbol values from actual input tensors
    3. Resolve shapes for any tensor/op

    Example symbolic_context:
    {
        "symbols": {
            "s0": {
                "name": "batch",
                "trace_value": 2,
                "min": 1, "max": 64,
                "source": "input::hidden_states::dim_0"
            }
        },
        "expressions": {
            "e0": "s0 * s1"
        }
    }
    """

    def __init__(
        self,
        symbolic_context: Optional[Dict[str, Any]] = None,
        strict: bool = False
    ):
        """
        Initialize resolver.

        Args:
            symbolic_context: The symbolic_context from graph.json.
                             If None or empty, resolver works in passthrough mode.
            strict: If True, raise error when using trace_value fallback.
                   ZERO FALLBACK: In strict mode, all symbols must be bound explicitly.
        """
        self._context = symbolic_context or {}
        self._symbols = self._context.get("symbols", {})
        self._expressions = self._context.get("expressions", {})
        self._strict = strict

        # Runtime values (bound from actual inputs)
        self._runtime_values: Dict[str, int] = {}
        self._bound = False

    @property
    def symbolic_shapes_enabled(self) -> bool:
        """Check if this graph has symbolic shapes."""
        return bool(self._symbols)

    def _get_nested_input(self, inputs: Dict[str, Any], key: str) -> Any:
        """
        Get value from inputs dict, supporting nested keys (dot notation).

        For key="added_cond_kwargs.resolution":
          - First checks if "added_cond_kwargs.resolution" exists directly
          - Then checks if "added_cond_kwargs" exists and has "resolution" key

        Args:
            inputs: Input dict (may contain nested dicts)
            key: Key to look up (may contain dots for nested access)

        Returns:
            The value if found, None otherwise
        """
        # Direct lookup first
        if key in inputs:
            return inputs[key]

        # Nested lookup for dotted names
        if "." in key:
            parts = key.split(".")
            value = inputs
            for part in parts:
                if isinstance(value, dict) and part in value:
                    value = value[part]
                else:
                    return None
            return value

        return None

    def bind_from_inputs(
        self,
        inputs: Dict[str, torch.Tensor],
        tensor_specs: Dict[str, Dict[str, Any]]
    ) -> None:
        """
        Bind symbol values from actual input tensors.

        Args:
            inputs: Dict of input_name -> tensor
            tensor_specs: Dict of tensor_id -> tensor spec from graph.json

        Raises:
            ShapeResolutionError: If symbol binding fails
        """
        if not self._symbols:
            self._bound = True
            return

        # Clear stale values — critical for AR loops where input shapes change
        # between iterations (e.g. seq_len grows from 1→2→3...).
        # Without clearing, old symbol values persist even when the source
        # input is no longer present, causing view/reshape to use wrong dims.
        self._runtime_values.clear()

        # For each symbol, find its source and bind
        for symbol_id, symbol_info in self._symbols.items():
            source = symbol_info.get("source", "")

            # Parse source: "input::hidden_states::dim_0"
            if source.startswith("input::"):
                parts = source.split("::")
                if len(parts) >= 3:
                    input_name = parts[1]
                    dim_str = parts[2]

                    # Extract dim index
                    dim_match = re.match(r"dim_(\d+)", dim_str)
                    if dim_match:
                        dim_idx = int(dim_match.group(1))

                        # Find the tensor (supports nested dicts like "added_cond_kwargs.resolution")
                        tensor = self._get_nested_input(inputs, input_name)
                        if tensor is not None:
                            if hasattr(tensor, 'ndim') and dim_idx < tensor.ndim:
                                value = tensor.shape[dim_idx]
                                self._bind_symbol(symbol_id, value, symbol_info)
                            elif hasattr(tensor, '__len__') and dim_idx < len(tensor):
                                # Handle list/tuple values (e.g., resolution=[1024, 1024])
                                value = tensor[dim_idx] if isinstance(tensor[dim_idx], int) else len(tensor)
                                self._bind_symbol(symbol_id, value, symbol_info)
                            elif hasattr(tensor, 'ndim'):
                                raise ShapeResolutionError(
                                    f"Symbol {symbol_id}: dim {dim_idx} out of range "
                                    f"for input '{input_name}' with shape {tensor.shape}"
                                )
                        else:
                            # Try to find by tensor_id pattern
                            tensor_id = f"input::{input_name}"
                            for name, t in inputs.items():
                                if name == input_name or tensor_id.endswith(name):
                                    if hasattr(t, 'ndim') and dim_idx < t.ndim:
                                        value = t.shape[dim_idx]
                                        self._bind_symbol(symbol_id, value, symbol_info)
                                        break
                            else:
                                logger.warning(
                                    f"Symbol {symbol_id}: Cannot find input '{input_name}' "
                                    f"in provided inputs: {list(inputs.keys())}"
                                )

        self._bound = True
        logger.debug(f"Bound symbols: {self._runtime_values}")

    def _bind_symbol(
        self,
        symbol_id: str,
        value: int,
        symbol_info: Dict[str, Any]
    ) -> None:
        """
        Bind a symbol to a value with constraint validation.

        Args:
            symbol_id: Symbol ID (e.g., "s0")
            value: Runtime value
            symbol_info: Symbol info with constraints

        Raises:
            ShapeResolutionError: If value violates constraints
        """
        # Validate constraints
        min_val = symbol_info.get("min", 0)
        max_val = symbol_info.get("max", float('inf'))

        if value < min_val:
            raise ShapeResolutionError(
                f"Symbol {symbol_id}={value} violates min constraint ({min_val})"
            )
        if value > max_val:
            raise ShapeResolutionError(
                f"Symbol {symbol_id}={value} violates max constraint ({max_val})"
            )

        self._runtime_values[symbol_id] = value
        logger.debug(f"Bound {symbol_id}={value} (name={symbol_info.get('name', '?')})")

    def resolve(self, shape: Any) -> Any:
        """
        Resolve a symbolic shape to concrete values.

        Args:
            shape: Can be:
                - List/tuple like ["s0", 4, 128, 128]
                - Single value like "s0" or 4
                - Expression like "e0"
                - Already concrete

        Returns:
            Resolved value (int, list, tuple)

        Raises:
            ShapeResolutionError: If resolution fails
        """
        if shape is None:
            return None

        # List/tuple: resolve each element
        if isinstance(shape, (list, tuple)):
            resolved = [self._resolve_single(dim) for dim in shape]
            return type(shape)(resolved)

        # Single value
        return self._resolve_single(shape)

    def _resolve_single(self, value: Any) -> Any:
        """
        Resolve a single dimension value.

        Handles SymInt objects and dict serialization format.
        Handles string symbol references for backward compatibility.
        """
        # Already concrete int/float
        if isinstance(value, (int, float)):
            return int(value)

        # SymInt object (in-memory)
        if SymInt is not None and isinstance(value, SymInt):
            return value.resolve(self._runtime_values)

        # Dict format (from JSON serialization)
        if isinstance(value, dict):
            return self._resolve_symint_dict(value)

        # String handling
        if isinstance(value, str):
            # Symbol reference (s0, s1, ...)
            if value in self._runtime_values:
                return self._runtime_values[value]

            # Check if it's a symbol we know but haven't bound
            if value in self._symbols:
                # STRICT MODE: No fallback allowed
                if self._strict:
                    raise ShapeResolutionError(
                        f"ZERO FALLBACK: Symbol '{value}' not bound at runtime (strict mode). "
                        f"Ensure all dynamic dimensions are resolved from actual inputs."
                    )

                # Non-strict: Use trace value as fallback (with warning)
                trace_val = self._symbols[value].get("trace_value")
                if trace_val is not None:
                    logger.warning(
                        f"Symbol {value} not bound, using trace_value={trace_val}"
                    )
                    return trace_val
                raise ShapeResolutionError(
                    f"Symbol {value} not bound and no trace_value available"
                )

            # Expression reference (e0, e1, ...)
            if value in self._expressions:
                return self._evaluate_expression(self._expressions[value])

            # Try to parse as int
            try:
                return int(value)
            except ValueError:
                pass

            # Check if it's an inline expression (e.g., "s0 * s1")
            if any(sym_id in value for sym_id in self._runtime_values):
                return self._evaluate_expression(value)

            # Unknown - return as-is (might be a string literal)
            return value

        # Unknown type - return as-is
        return value

    def _resolve_symint_dict(self, data: Dict[str, Any]) -> int:
        """
        Resolve a SymInt serialized as a dict.

        Handles:
        - {"type": "symbol", "id": "s0", "trace": 2}
        - {"type": "mul", "left": {...}, "right": {...}, "trace": 240}
        - etc.
        """
        type_str = data.get("type", "")

        # Constant
        if type_str == "const" or "value" in data:
            return data.get("value", data.get("trace", 0))

        # Symbol reference
        if type_str == "symbol":
            # Handle both formats: (id/trace) and graph format (symbol_id/trace_value)
            symbol_id = data.get("id") or data.get("symbol_id")
            if symbol_id in self._runtime_values:
                return self._runtime_values[symbol_id]
            # Fallback to trace value (both formats)
            trace = data.get("trace") if data.get("trace") is not None else data.get("trace_value")
            if trace is not None:
                if self._strict:
                    raise ShapeResolutionError(
                        f"ZERO FALLBACK: Symbol '{symbol_id}' not bound (strict mode)"
                    )
                logger.warning(f"Symbol {symbol_id} not bound, using trace={trace}")
                return trace
            raise ShapeResolutionError(f"Symbol {symbol_id} not bound")

        # Unary: neg
        if type_str == "neg":
            operand = self._resolve_symint_dict(data["operand"])
            return -operand

        # Binary operations
        if type_str in ("add", "sub", "mul", "floordiv", "mod"):
            left = self._resolve_single(data["left"])
            right = self._resolve_single(data["right"])

            if type_str == "add":
                return left + right
            elif type_str == "sub":
                return left - right
            elif type_str == "mul":
                return left * right
            elif type_str == "floordiv":
                return left // right
            elif type_str == "mod":
                return left % right

        # Product: multiply factors together (e.g., s1 * s2)
        if type_str == "product":
            factors = data.get("factors", [])
            if not factors:
                return data.get("trace_value", 0)
            result = 1
            for factor in factors:
                result *= self._resolve_single(factor)
            return result

        # Scaled product: (scale_h * factors[0]) * (scale_w * factors[1])
        # Used by VAE expand operations that scale height/width
        if type_str == "scaled_product":
            factors = data.get("factors", [])
            scale_h = data.get("scale_h", 1)
            scale_w = data.get("scale_w", 1)

            if len(factors) >= 2:
                h_val = self._resolve_single(factors[0])
                w_val = self._resolve_single(factors[1])
                return (scale_h * h_val) * (scale_w * w_val)
            elif len(factors) == 1:
                val = self._resolve_single(factors[0])
                return scale_h * scale_w * val
            else:
                return data.get("trace_value", 0)

        # Scaled symbol: scale * symbol_value
        if type_str == "scaled_symbol":
            symbol_id = data.get("symbol_id") or data.get("id")
            scale = data.get("scale", 1)
            if symbol_id in self._runtime_values:
                return scale * self._runtime_values[symbol_id]
            # Fallback to trace
            trace = data.get("trace_value") or data.get("trace")
            if trace is not None:
                return trace
            raise ShapeResolutionError(f"Scaled symbol {symbol_id} not bound")

        raise ShapeResolutionError(f"Unknown SymInt type: {type_str}")

    def _evaluate_expression(self, expr: str) -> int:
        """
        Evaluate a symbolic expression.

        Args:
            expr: Expression string like "s0 * s1" or "s1 / 2"

        Returns:
            Evaluated integer result
        """
        # Replace symbols with values
        resolved_expr = expr
        for symbol_id, value in self._runtime_values.items():
            resolved_expr = resolved_expr.replace(symbol_id, str(value))

        # Safe evaluation (only arithmetic via AST)
        try:
            import ast
            import operator
            allowed_chars = set("0123456789+-*/() ")
            if not all(c in allowed_chars for c in resolved_expr):
                raise ShapeResolutionError(
                    f"Unsafe expression: {expr} -> {resolved_expr}"
                )
            if len(resolved_expr) > 256:
                raise ShapeResolutionError(
                    f"Expression too long ({len(resolved_expr)} chars): {expr}"
                )
            _safe_ops = {
                ast.Add: operator.add, ast.Sub: operator.sub,
                ast.Mult: operator.mul, ast.FloorDiv: operator.floordiv,
                ast.Div: operator.truediv, ast.USub: operator.neg,
            }
            def _safe_eval(node):
                if isinstance(node, ast.Constant) and isinstance(node.value, (int, float)):
                    return node.value
                if isinstance(node, ast.BinOp) and type(node.op) in _safe_ops:
                    return _safe_ops[type(node.op)](_safe_eval(node.left), _safe_eval(node.right))
                if isinstance(node, ast.UnaryOp) and type(node.op) in _safe_ops:
                    return _safe_ops[type(node.op)](_safe_eval(node.operand))
                raise ShapeResolutionError(f"Unsafe AST node: {ast.dump(node)}")
            tree = ast.parse(resolved_expr, mode='eval')
            result = _safe_eval(tree.body)
            return int(result)
        except Exception as e:
            raise ShapeResolutionError(
                f"Failed to evaluate expression '{expr}': {e}"
            )

    def resolve_tensor_shape(
        self,
        tensor_spec: Dict[str, Any]
    ) -> Tuple[int, ...]:
        """
        Resolve shape for a tensor spec from graph.json.

        Handles SymInt dict format (expression trees).
        Handles string symbol references.

        Args:
            tensor_spec: Tensor spec with "shape" or "shape_concrete"

        Returns:
            Concrete shape tuple
        """
        if "shape" in tensor_spec:
            shape = tensor_spec["shape"]
            # Check if shape contains symbolic elements
            has_symbolic = False
            for d in shape:
                if isinstance(d, str):
                    has_symbolic = True
                    break
                if isinstance(d, dict):
                    # SymInt dict format
                    has_symbolic = True
                    break

            if has_symbolic:
                return tuple(self.resolve(shape))
            # All concrete ints
            return tuple(shape)

        # Fallback: Use shape_concrete or shape
        if "shape_concrete" in tensor_spec:
            return tuple(tensor_spec["shape_concrete"])

        raise ShapeResolutionError(
            f"Tensor spec missing shape: {tensor_spec.get('tensor_id', '?')}"
        )

    def get_bound_symbols(self) -> Dict[str, int]:
        """Get all bound symbol values."""
        return dict(self._runtime_values)

    def __repr__(self) -> str:
        return (
            f"SymbolicShapeResolver("
            f"symbols={len(self._symbols)}, "
            f"bound={len(self._runtime_values)})"
        )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import functools

import jax
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu

from vllm.utils.math_utils import cdiv


def _kv_cache_update_kernel(
    # Prefetch
    slices_ref,  # [3, padded_num_slices], list of (kv_cache_start,
    # new_kv_start, slice_len)
    num_slices_ref,  # [1]
    # Input
    new_kv_hbm_ref,  # [num_tokens, num_combined_kv_heads, head_dim]
    kv_cache_hbm_ref,  # [total_num_pages * page_size, num_combined_kv_heads,
    # head_dim]
    # Output
    _,  # [total_num_pages * page_size, num_combined_kv_heads, head_dim]
    # Scratch
    scratch,  # [num_slices_per_block, page_size, num_combined_kv_heads,
    # head_dim]
    sem,
):
    async_copies = []
    block_idx = pl.program_id(0)
    num_slices_per_block = scratch.shape[0]

    # Copy from new_kv_hbm_ref to scratch
    for i in range(num_slices_per_block):
        offset_i = i + block_idx * num_slices_per_block
        new_kv_start = jax.lax.select(
            offset_i < num_slices_ref[0], slices_ref[1, offset_i], 0
        )
        length = jax.lax.select(
            offset_i < num_slices_ref[0], slices_ref[2, offset_i], 0
        )
        async_copy = pltpu.make_async_copy(
            new_kv_hbm_ref.at[pl.ds(new_kv_start, length), ...],
            scratch.at[i, pl.ds(0, length), ...],
            sem,
        )
        async_copy.start()
        async_copies.append(async_copy)

    for async_copy in async_copies:
        async_copy.wait()

    # Copy from scratch to kv_cache_hbm_ref
    async_copies.clear()
    for i in range(num_slices_per_block):
        offset_i = i + block_idx * num_slices_per_block
        kv_cache_start = jax.lax.select(
            offset_i < num_slices_ref[0], slices_ref[0, offset_i], 0
        )
        length = jax.lax.select(
            offset_i < num_slices_ref[0], slices_ref[2, offset_i], 0
        )
        async_copy = pltpu.make_async_copy(
            scratch.at[i, pl.ds(0, length), ...],
            kv_cache_hbm_ref.at[pl.ds(kv_cache_start, length), ...],
            sem,
        )
        async_copy.start()
        async_copies.append(async_copy)
    for async_copy in async_copies:
        async_copy.wait()


@functools.partial(
    jax.jit,
    static_argnames=["page_size", "num_slices_per_block"],
)
def kv_cache_update(
    # [total_num_token, num_combined_kv_heads, head_dim]
    new_kv: jax.Array,
    # [3, slices], list of (kv_cache_start, new_kv_start, slice_len)
    slices: jax.Array,
    # [total_num_pages * page_size, num_combined_kv_heads, head_dim]
    kv_cache: jax.Array,
    # [1]
    num_kv_update_slices: jax.Array,
    *,
    page_size: int = 32,
    num_slices_per_block: int = 8,
):
    _, num_combined_kv_heads, head_dim = new_kv.shape
    assert kv_cache.shape[1] == num_combined_kv_heads
    assert kv_cache.shape[2] == head_dim
    assert head_dim % 128 == 0
    # TODO: Add dynamic check to make sure that the all the slice lengths are
    # smaller or equal to page_size

    in_specs = [
        pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
        pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY),
    ]

    out_specs = [pl.BlockSpec(memory_space=pltpu.TPUMemorySpace.ANY)]
    out_shape = [jax.ShapeDtypeStruct(kv_cache.shape, dtype=kv_cache.dtype)]

    scalar_prefetches = [slices, num_kv_update_slices]
    scratch = pltpu.VMEM(
        (num_slices_per_block, page_size, num_combined_kv_heads, head_dim),
        new_kv.dtype,
    )

    scratch_shapes = [
        scratch,
        pltpu.SemaphoreType.DMA,
    ]

    kernel = pl.pallas_call(
        _kv_cache_update_kernel,
        grid_spec=pltpu.PrefetchScalarGridSpec(
            num_scalar_prefetch=len(scalar_prefetches),
            in_specs=in_specs,
            out_specs=out_specs,
            grid=(cdiv(num_kv_update_slices[0], num_slices_per_block),),
            scratch_shapes=scratch_shapes,
        ),
        out_shape=out_shape,
        input_output_aliases={len(scalar_prefetches) + 1: 0},
    )

    return kernel(*scalar_prefetches, new_kv, kv_cache)[0]



import logging

import torch
import triton
import triton.language as tl

from flag_gems.utils import pointwise_dynamic
from flag_gems.utils.triton_lang_extension import div_rn, div_rz, fmod, trunc

logger = logging.getLogger(__name__)


@pointwise_dynamic(promotion_methods=[(0, 1, "INT_TO_FLOAT")])
@triton.jit
def true_div_func(x, y):
    return x / y


@pointwise_dynamic(is_tensor=[True, False], promotion_methods=[(0, 1, "INT_TO_FLOAT")])
@triton.jit
def true_div_func_tensor_scalar(x, y):
    return x / y


@pointwise_dynamic(is_tensor=[False, True], promotion_methods=[(0, 1, "INT_TO_FLOAT")])
@triton.jit
def true_div_func_scalar_tensor(x, y):
    return x / y


def true_divide(A, B):
    logger.debug("GEMS TRUE_DIVIDE")
    if isinstance(A, torch.Tensor) and isinstance(B, torch.Tensor):
        return true_div_func(A, B)
    elif isinstance(A, torch.Tensor):
        return true_div_func_tensor_scalar(A, B)
    elif isinstance(B, torch.Tensor):
        return true_div_func_scalar_tensor(A, B)
    else:
        # Both scalar
        return torch.tensor(A / B)


def true_divide_out(A, B, out):
    logger.debug("GEMS TRUE_DIVIDE OUT")
    if isinstance(A, torch.Tensor) and isinstance(B, torch.Tensor):
        return true_div_func(A, B, out0=out)
    elif isinstance(A, torch.Tensor):
        return true_div_func_tensor_scalar(A, B, out0=out)
    elif isinstance(B, torch.Tensor):
        return true_div_func_scalar_tensor(A, B, out0=out)
    else:
        # Both scalar
        return torch.tensor(A / B) if out is None else out.fill_(A / B)


def true_divide_(A, B):
    logger.debug("GEMS TRUE_DIVIDE_")
    if isinstance(B, torch.Tensor):
        return true_div_func(A, B, out0=A)
    else:
        return true_div_func_tensor_scalar(A, B, out0=A)


@pointwise_dynamic(promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def trunc_div_func(x, y):
    return trunc(div_rz(x, y))


@pointwise_dynamic(is_tensor=[True, False], promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def trunc_div_func_tensor_scalar(x, y):
    return trunc(div_rz(x, y))


@pointwise_dynamic(is_tensor=[False, True], promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def trunc_div_func_scalar_tensor(x, y):
    return trunc(div_rz(x, y))


def trunc_divide(A, B):
    logger.debug("GEMS TRUNC_DIVIDE")
    if isinstance(A, torch.Tensor) and isinstance(B, torch.Tensor):
        return trunc_div_func(A, B)
    elif isinstance(A, torch.Tensor):
        return trunc_div_func_tensor_scalar(A, B)
    elif isinstance(B, torch.Tensor):
        return trunc_div_func_scalar_tensor(A, B)
    else:
        # Both scalar
        return torch.tensor(A / B)


def trunc_divide_(A, B):
    logger.debug("GEMS TRUNC_DIVIDE_")
    if isinstance(B, torch.Tensor):
        return trunc_div_func(A, B, out0=A)
    else:
        return trunc_div_func_tensor_scalar(A, B, out0=A)


@triton.jit
def _int_floordiv(x, y):
    # TODO: request Triton to add an integer remainder builtin
    # The semantic of Triton floordiv differs from Pytorch/Numpy
    # Triton floordiv equates to
    #     (x - np.fmod(x, y)) / y
    # whereas Pytorch floordiv is
    #     (x - np.remainder(x, y)) y
    # The results show a one off difference when
    #     C1) x and y have opposite signs
    # and C2) x is not multiples of y.
    # Apart from the above, there's an erroneous case x // 0 returns -1
    # whereas in Pytorch x // 0 returns -1 if x >=0 and -2 if x < 0
    # but this special case is coalesced into the c1 and c2 check so
    # there's extra handling.
    r = x % y
    c1 = r != 0
    c2 = (x < 0) ^ (y < 0)
    return tl.where(c1 & c2, x // y - 1, x // y)


# TO be consistent with python, numpy and torch, we have to implement it in the
# following way.
# CPython
# https://github.com/python/cpython/blob/ace008c531dd685a30c1dd68f9b5ba35f20171cf/Objects/floatobject.c#L636
# numpy
# https://github.com/numpy/numpy/blob/a4ad142aa1282a77bbb05acd706cb57c9cc29846/numpy/_core/src/npymath/npy_math_internal.h.src#L532
# torch
# https://github.com/pytorch/pytorch/blob/d6d9183456cd07ca0b361a194b98c2fb196e7c36/c10/util/generic_math.h#L23
@triton.jit
def _float_floordiv(x, y):
    # NOTE: fmod's sign is the same as the dividend
    remainder = fmod(x, y)
    imperfect = remainder != 0.0
    different_sign = (x < 0) ^ (y < 0)

    # NOTE: we have to use div_rn explicitly here
    q = div_rn(x - remainder, y)
    q = tl.where(imperfect & different_sign, q - 1, q)

    floor_q = tl.math.floor(q)
    c = q - floor_q > 0.5
    floor_q = tl.where(c, floor_q + 1.0, floor_q)

    q_is_zeros = q == 0.0
    floor_q = tl.where(q_is_zeros, tl.where(different_sign, -0.0, 0.0), floor_q)

    is_div_by_zero = y == 0.0
    float_division = x / y
    out = tl.where(is_div_by_zero, float_division, floor_q)
    return out


@pointwise_dynamic(promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def floor_div_func(x, y):
    if x.type.scalar.is_int() & x.type.scalar.is_int():
        return _int_floordiv(x, y)
    else:
        return _float_floordiv(x, y)


@pointwise_dynamic(is_tensor=[True, False], promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def floor_div_func_tensor_scalar(x, y):
    if x.type.scalar.is_int() & x.type.scalar.is_int():
        return _int_floordiv(x, y)
    else:
        return _float_floordiv(x, y)


@pointwise_dynamic(is_tensor=[False, True], promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def floor_div_func_scalar_tensor(x, y):
    if x.type.scalar.is_int() & x.type.scalar.is_int():
        return _int_floordiv(x, y)
    else:
        return _float_floordiv(x, y)


def floor_divide(A, B):
    logger.debug("GEMS FLOOR_DIVIDE")
    if isinstance(A, torch.Tensor) and isinstance(B, torch.Tensor):
        return floor_div_func(A, B)
    elif isinstance(A, torch.Tensor):
        return floor_div_func_tensor_scalar(A, B)
    elif isinstance(B, torch.Tensor):
        return floor_div_func_scalar_tensor(A, B)
    else:
        # Both scalar
        return torch.tensor(A // B)


def floor_divide_(A, B):
    logger.debug("GEMS FLOOR_DIVIDE_")
    if isinstance(B, torch.Tensor):
        return floor_div_func(A, B, out0=A)
    else:
        return floor_div_func_tensor_scalar(A, B, out0=A)


def div_mode(A, B, rounding_mode=None):
    if rounding_mode is None:
        return true_divide(A, B)
    elif rounding_mode == "trunc":
        return trunc_divide(A, B)
    elif rounding_mode == "floor":
        return floor_divide(A, B)
    else:
        msg = f"div expected rounding_mode to be one of None, 'trunc', or 'floor' but found {rounding_mode}."
        raise ValueError(msg)


def div_mode_(A, B, rounding_mode=None):
    if rounding_mode is None:
        return true_divide_(A, B)
    elif rounding_mode == "trunc":
        return trunc_divide_(A, B)
    elif rounding_mode == "floor":
        return floor_divide_(A, B)
    else:
        msg = f"div expected rounding_mode to be one of None, 'trunc', or 'floor' but found {rounding_mode}."
        raise ValueError(msg)


@triton.jit
def _remainder(x, y):
    r = x % y
    c1 = r != 0
    c2 = (x < 0) ^ (y < 0)
    return tl.where(c1 & c2, r + y, r)


@pointwise_dynamic(promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def rem_tt(x, y):
    return _remainder(x, y)


@pointwise_dynamic(is_tensor=[True, False], promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def rem_ts(x, y):
    return _remainder(x, y)


@pointwise_dynamic(is_tensor=[False, True], promotion_methods=[(0, 1, "DEFAULT")])
@triton.jit
def rem_st(x, y):
    return _remainder(x, y)


def remainder(A, B):
    logger.debug("GEMS REMAINDER")
    if isinstance(A, torch.Tensor) and isinstance(B, torch.Tensor):
        return rem_tt(A, B)
    elif isinstance(A, torch.Tensor):
        return rem_ts(A, B)
    elif isinstance(B, torch.Tensor):
        return rem_st(A, B)
    else:
        # Both scalar
        return torch.tensor(A % B)


def remainder_(A, B):
    logger.debug("GEMS REMAINDER_")
    if isinstance(B, torch.Tensor):
        return rem_tt(A, B, out0=A)
    else:
        return rem_ts(A, B, out0=A)



import logging
import math

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry, libtuner
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger(__name__)


@libentry()
@libtuner(
    configs=runtime.get_tuned_config("bmm"),
    key=["M", "N", "K"],
)
@triton.heuristics(runtime.get_heuristic_config("bmm"))
@triton.heuristics(
    {
        "UPGRADE": lambda args: math.ceil(
            (args["M"] * args["N"] * args["batch"]) / (args["TILE_M"] * args["TILE_M"])
        ).bit_length()
        > 32,
    }
)
@triton.jit
def bmm_kernel(
    A,
    B,
    O,
    M,
    N,
    K,
    batch,
    TILE_M: tl.constexpr,
    TILE_N: tl.constexpr,
    TILE_K: tl.constexpr,
    GROUP_M: tl.constexpr,
    DIVISIBLE_M: tl.constexpr,
    DIVISIBLE_N: tl.constexpr,
    DIVISIBLE_K: tl.constexpr,
    UPGRADE: tl.constexpr,
):
    # batch offsets
    pid_b = tle.program_id(2)
    A += pid_b * M * K
    B += pid_b * K * N
    O += pid_b * M * N

    if UPGRADE:
        pidx = tle.program_id(0)
        pidy = tle.program_id(1)
    else:
        pidx = tl.program_id(0)
        pidy = tl.program_id(1)

    if GROUP_M == 1:
        pid_m, pid_n = pidx, pidy
    else:
        # reorder CTAs
        if UPGRADE:
            gridx = tle.num_programs(0)
            gridy = tle.num_programs(1)
        else:
            gridx = tl.num_programs(0)
            gridy = tl.num_programs(1)
        pid = pidx + pidy * gridx

        num_CTA_per_group = gridy * GROUP_M

        group_id = pid // num_CTA_per_group
        inner_group_id = pid % num_CTA_per_group
        GROUP_SIZE = tl.where(
            (group_id * GROUP_M + GROUP_M) > gridx, gridx % GROUP_M, GROUP_M
        )
        pid_m = group_id * GROUP_M + inner_group_id % GROUP_SIZE
        pid_n = inner_group_id // GROUP_SIZE

    offs_m = pid_m * TILE_M + tl.arange(0, TILE_M)
    offs_n = pid_n * TILE_N + tl.arange(0, TILE_N)
    offs_k = tl.arange(0, TILE_K)

    if not DIVISIBLE_M:
        mask_m = offs_m < M
    if not DIVISIBLE_N:
        mask_n = offs_n < N

    a_ptrs = A + offs_m[:, None] * K + offs_k[None, :]
    b_ptrs = B + offs_k[:, None] * N + offs_n[None, :]
    o_ptrs = O + offs_m[:, None] * N + offs_n[None, :]

    num_iters = tl.cdiv(K, TILE_K)
    o = tl.zeros((TILE_M, TILE_N), dtype=tl.float32)
    for _ in range(num_iters):
        if DIVISIBLE_K:
            if DIVISIBLE_M:
                mask_a = None
            else:
                mask_a = mask_m[:, None]
            if DIVISIBLE_N:
                mask_b = None
            else:
                mask_b = mask_n[None, :]
        else:
            mask_k = offs_k < K
            if DIVISIBLE_M:
                mask_a = mask_k[None, :]
            else:
                mask_a = mask_m[:, None] & mask_k[None, :]
            if DIVISIBLE_N:
                mask_b = mask_k[:, None]
            else:
                mask_b = mask_k[:, None] & mask_n[None, :]

        a = tl.load(a_ptrs, mask_a)
        b = tl.load(b_ptrs, mask_b)

        offs_k += TILE_K
        a_ptrs += TILE_K
        b_ptrs += TILE_K * N

        o += tl.dot(a, b, allow_tf32=False)

    if DIVISIBLE_M and DIVISIBLE_N:
        mask_c = None
    elif DIVISIBLE_M and not DIVISIBLE_N:
        mask_c = mask_n[None, :]
    elif not DIVISIBLE_M and DIVISIBLE_N:
        mask_c = mask_m[:, None]
    else:
        mask_c = mask_m[:, None] & mask_n[None, :]
    tl.store(o_ptrs, o, mask_c)


def bmm(A, B):
    logger.debug("METAX GEMS BMM")
    batch, M, K = A.shape
    _, _, N = B.shape
    A = A.contiguous()
    B = B.contiguous()
    out = torch.empty((batch, M, N), dtype=A.dtype, device=A.device)

    grid_fn = lambda meta: (
        triton.cdiv(meta["M"], meta["TILE_M"]),
        triton.cdiv(meta["N"], meta["TILE_N"]),
        batch,
    )
    with torch_device_fn.device(A.device):
        bmm_kernel[grid_fn](A, B, out, M, N, K, batch)
    return out



import logging
import math

import torch
import triton
import triton.language as tl

from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry, libtuner
from flag_gems.utils.shape_utils import volume

from ..utils import TOTAL_CORE_NUM

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@libentry()
@libtuner(
    configs=[
        triton.Config(kwargs={"BLOCK_SIZE": 1024}, num_stages=1, num_warps=1),
        triton.Config(kwargs={"BLOCK_SIZE": 4096}, num_stages=1, num_warps=1),
        triton.Config(kwargs={"BLOCK_SIZE": 16384}, num_stages=1, num_warps=1),
        triton.Config(kwargs={"BLOCK_SIZE": 65536}, num_stages=1, num_warps=1),
    ],
    key=["n_elements"],
)
@triton.jit(do_not_specialize=["fill_value_or_ptr"])
def full_tensor_kernel(
    output_ptr,
    n_elements,
    fill_value_ptr,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tl.program_id(axis=0)
    num_jobs = tl.num_programs(axis=0)
    block_start = pid * BLOCK_SIZE
    step = num_jobs * BLOCK_SIZE
    block_start = block_start.to(tl.int64)
    for block_start_offset in range(block_start, n_elements, step):
        offsets = block_start_offset + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        fill_value = tl.load(fill_value_ptr)
        tl.store(output_ptr + offsets, fill_value, mask=mask)


@libentry()
@libtuner(
    configs=[
        triton.Config(kwargs={"BLOCK_SIZE": 1024}, num_stages=1, num_warps=1),
        triton.Config(kwargs={"BLOCK_SIZE": 4096}, num_stages=1, num_warps=1),
        triton.Config(kwargs={"BLOCK_SIZE": 16384}, num_stages=1, num_warps=1),
        triton.Config(kwargs={"BLOCK_SIZE": 65536}, num_stages=1, num_warps=1),
    ],
    key=["n_elements"],
)
@triton.jit(do_not_specialize=["fill_value_or_ptr"])
def full_scalar_kernel(
    output_ptr,
    n_elements,
    fill_value,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tl.program_id(axis=0)
    num_jobs = tl.num_programs(axis=0)
    block_start = pid * BLOCK_SIZE
    step = num_jobs * BLOCK_SIZE
    block_start = block_start.to(tl.int64)
    for block_start_offset in range(block_start, n_elements, step):
        offsets = block_start_offset + tl.arange(0, BLOCK_SIZE)
        mask = offsets < n_elements
        tl.store(output_ptr + offsets, fill_value, mask=mask)


ALL_INT_DTYPES = (torch.int8, torch.int16, torch.int32, torch.int64)
ALL_FLOAT_DTYPES = (torch.bfloat16, torch.float16, torch.float32, torch.float64)


def check_dtype(fill_value, dtype, device):
    if isinstance(fill_value, bool):
        if dtype != torch.bool:
            fill_value = int(fill_value)
    else:
        if isinstance(fill_value, float) and math.isinf(fill_value):
            if dtype not in ALL_FLOAT_DTYPES:
                raise RuntimeError(
                    f"value {fill_value!r} cannot be converted to type {dtype} without overflow"
                )
        elif (
            dtype in ALL_INT_DTYPES
            and (
                fill_value < torch.iinfo(dtype).min
                or fill_value > torch.iinfo(dtype).max
            )
        ) or (
            dtype in ALL_FLOAT_DTYPES
            and (
                fill_value < torch.finfo(dtype).min
                or fill_value > torch.finfo(dtype).max
            )
        ):
            raise RuntimeError(
                f"value cannot be converted to type {dtype} without overflow"
            )
    if dtype in ALL_FLOAT_DTYPES:
        fill_value = torch.tensor(fill_value, dtype=dtype, device=device)
    return fill_value


def full(size, fill_value, *, dtype=None, layout=None, device=None, pin_memory=None):
    logger.debug("GEMS_CAMBRICON FULL")
    if device is None:
        device = torch.device("cpu")
    if dtype is None:
        if isinstance(fill_value, bool):
            dtype = torch.bool
        elif isinstance(fill_value, int):
            dtype = torch.int64
        else:
            dtype = torch.get_default_dtype()
    else:
        fill_value = check_dtype(fill_value, dtype, device)

    out = torch.empty(size, device=device, dtype=dtype)
    N = volume(size)
    grid_fn = lambda meta: (min(triton.cdiv(N, meta["BLOCK_SIZE"]), TOTAL_CORE_NUM),)
    with torch_device_fn.device(device):
        if isinstance(fill_value, torch.Tensor):
            full_tensor_kernel[grid_fn](
                out,
                N,
                fill_value,
            )
        else:
            full_scalar_kernel[grid_fn](
                out,
                N,
                fill_value,
            )
    return out



import logging

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry

from ..utils import TOTAL_CORE_NUM

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@libentry()
@triton.autotune(
    configs=runtime.get_tuned_config("nonzero"),
    key=[
        "n_elements",
    ],
)
@triton.jit
def nonzero_kernel(
    inp,
    prefix_sum,
    out,
    n_elements,
    shape,
    ndim: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tl.program_id(0)
    num_jobs = tl.num_programs(0)
    block_start = pid * BLOCK_SIZE
    step = num_jobs * BLOCK_SIZE
    for block_start_offset in range(block_start, n_elements, step):
        offset = block_start_offset + tl.arange(0, BLOCK_SIZE)
        mask = offset < n_elements

        inp_vals = tl.load(inp + offset, mask=mask, other=0.0).to(tl.int1)
        nonzero_mask = mask and inp_vals
        out_row_offset = tl.load(prefix_sum + offset, mask=nonzero_mask) - 1
        out_col_offset = tl.arange(0, ndim)
        out_offsets = out_row_offset[:, None] * ndim + out_col_offset[None, :]
        out_vals = tl.zeros((BLOCK_SIZE, ndim), tl.int32)
        idx_flat = offset
        for dim in range(ndim - 1, -1, -1):
            dim_size = tl.load(shape + dim)
            remainder = idx_flat % dim_size
            idx_flat //= dim_size
            out_vals[:, dim] = remainder
        tl.store(out + out_offsets.to(tl.int32), out_vals, mask=nonzero_mask[:, None])


def nonzero(inp, *, as_tuple=False):
    logger.debug("GEMS_CAMBRICON NONZERO")

    inp_ndim = inp.ndim

    inp = inp.contiguous()
    n_elements = inp.numel()
    inp_view = inp.view(n_elements)

    shape = torch.tensor(inp.shape, dtype=torch.int32, device=inp.device)

    inp_bool = inp_view
    if inp_view.dtype != torch.bool:
        inp_bool = inp_view != 0

    prefix_sum = inp_bool.cumsum(axis=0)
    num_nonzeros = n_elements
    out = torch.empty(num_nonzeros, inp_ndim, dtype=torch.int64, device=inp.device)
    grid = lambda meta: (
        min(triton.cdiv(n_elements, meta["BLOCK_SIZE"]), TOTAL_CORE_NUM),
    )
    with torch_device_fn.device(inp.device):
        nonzero_kernel[grid](inp_bool, prefix_sum, out, n_elements, shape, inp_ndim)

    num_nonzeros = prefix_sum[n_elements - 1].item()
    out = out[0:num_nonzeros]

    if as_tuple:
        return torch.unbind(out, dim=0)
    else:
        return out



"""
NeuroBrix Memory Management for I/O.

Provides:
1. IOConfig - Configuration for I/O operations
2. PinnedMemoryManager - Pinned memory for PCIe DMA transfers
3. PrefetchQueue - Overlapping I/O and compute

ZERO HARDCODE: Configuration from environment variables
"""

import os
import threading
from dataclasses import dataclass, field
from pathlib import Path
from queue import Queue, Empty
from typing import Dict, List, Optional, Any

import torch


# ============================================================================
# Configuration
# ============================================================================

@dataclass
class IOConfig:
    """I/O configuration for loaders."""

    # Number of parallel workers for disk I/O
    num_workers: int = 8

    # Use pinned memory for DMA transfers
    use_pinned_memory: bool = True

    # Enable prefetching (load next component while current executes)
    enable_prefetch: bool = True

    # Prefetch queue size (number of components to prefetch)
    prefetch_queue_size: int = 2

    # Log performance metrics
    log_performance: bool = True

    @classmethod
    def from_env(cls) -> "IOConfig":
        """Create config from environment variables."""
        return cls(
            num_workers=int(os.environ.get("NBX_IO_WORKERS", "8")),
            use_pinned_memory=os.environ.get("NBX_PINNED_MEMORY", "1") == "1",
            enable_prefetch=os.environ.get("NBX_PREFETCH", "1") == "1",
            prefetch_queue_size=int(os.environ.get("NBX_PREFETCH_SIZE", "2")),
            log_performance=os.environ.get("NBX_IO_LOG", "1") == "1",
        )


# Global config instance
_io_config: Optional[IOConfig] = None


def get_io_config() -> IOConfig:
    """Get global I/O configuration."""
    global _io_config
    if _io_config is None:
        _io_config = IOConfig.from_env()
    return _io_config


def set_io_config(config: IOConfig) -> None:
    """Set global I/O configuration."""
    global _io_config
    _io_config = config


# ============================================================================
# Pinned Memory Manager
# ============================================================================

class PinnedMemoryManager:
    """
    Manages pinned memory buffers for DMA transfers.

    Pinned memory enables PCIe DMA (16GB/s) instead of CPU copy (~8GB/s).

    Usage:
        manager = PinnedMemoryManager()
        pinned_dict = manager.to_pinned(tensors)
        gpu_dict = manager.to_device(pinned_dict, "cuda:0", non_blocking=True)
    """

    def __init__(self, enabled: bool = True):
        self.enabled = enabled and torch.cuda.is_available()
        self._stats = {"pinned_bytes": 0, "transfers": 0}

    def to_pinned(self, tensors: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Move tensors to pinned memory for fast GPU transfer.

        Args:
            tensors: Dict of name -> tensor (on CPU)

        Returns:
            Dict of name -> tensor (pinned memory)
        """
        if not self.enabled:
            return tensors

        pinned = {}
        for name, tensor in tensors.items():
            if tensor.device.type == "cpu" and not tensor.is_pinned():
                # Create pinned tensor and copy data
                pinned_tensor = torch.empty_like(tensor, pin_memory=True)
                pinned_tensor.copy_(tensor)
                pinned[name] = pinned_tensor
                self._stats["pinned_bytes"] += tensor.numel() * tensor.element_size()
            else:
                pinned[name] = tensor

        return pinned

    def to_device(
        self,
        tensors: Dict[str, torch.Tensor],
        device: str,
        non_blocking: bool = True,
    ) -> Dict[str, torch.Tensor]:
        """
        Transfer tensors to GPU device using DMA (non-blocking).

        Args:
            tensors: Dict of name -> tensor (pinned memory)
            device: Target device (e.g., "cuda:0")
            non_blocking: Use async transfer (True for DMA)

        Returns:
            Dict of name -> tensor (on GPU)
        """
        result = {}
        for name, tensor in tensors.items():
            result[name] = tensor.to(device, non_blocking=non_blocking)
            self._stats["transfers"] += 1
        return result

    def get_stats(self) -> Dict[str, Any]:
        """Get pinned memory statistics."""
        return {
            **self._stats,
            "pinned_mb": self._stats["pinned_bytes"] / 1e6,
        }


# ============================================================================
# Prefetch Queue
# ============================================================================

@dataclass
class PrefetchItem:
    """Item in the prefetch queue."""
    component_name: str
    shard_paths: List[Path]
    dtype: Optional[torch.dtype] = None
    device: str = "cpu"  # Prefetch to CPU/pinned, NOT GPU
    tensors: Optional[Dict[str, torch.Tensor]] = None
    ready: threading.Event = field(default_factory=threading.Event)


class PrefetchQueue:
    """
    Prefetch queue for overlapping I/O and compute.

    While GPU executes Component A, this loads Component B to pinned RAM.

    Usage:
        queue = PrefetchQueue()
        queue.start()

        # Queue next component for prefetch
        queue.prefetch("transformer", shard_paths, dtype)

        # Get prefetched weights (blocks if not ready)
        weights = queue.get("transformer", device="cuda:0")

        queue.stop()

    CRITICAL: Prefetch happens to CPU/pinned RAM only, never to GPU.
    This avoids VRAM OOM from holding two components simultaneously.
    """

    def __init__(self, config: Optional[IOConfig] = None):
        self.config = config or get_io_config()
        # Lazy import to avoid circular dependency
        self._loader = None

        self._queue: Queue[PrefetchItem] = Queue(
            maxsize=self.config.prefetch_queue_size
        )
        self._cache: Dict[str, PrefetchItem] = {}
        self._worker: Optional[threading.Thread] = None
        self._running = False
        self._lock = threading.Lock()

    def _get_loader(self):
        """Lazy load SmartLoader to avoid circular import."""
        if self._loader is None:
            from neurobrix.core.io.loader import SmartLoader
            self._loader = SmartLoader(self.config)
        return self._loader

    def start(self) -> None:
        """Start the prefetch worker thread."""
        if self._worker is not None:
            return

        self._running = True
        self._worker = threading.Thread(target=self._worker_loop, daemon=True)
        self._worker.start()


    def stop(self) -> None:
        """Stop the prefetch worker thread."""
        self._running = False
        if self._worker is not None:
            self._worker.join(timeout=5.0)
            self._worker = None

    def prefetch(
        self,
        component_name: str,
        shard_paths: List[Path],
        dtype: Optional[torch.dtype] = None,
    ) -> None:
        """
        Queue a component for prefetch.

        IMPORTANT: Prefetches to CPU/pinned RAM, not GPU.

        Args:
            component_name: Name of component to prefetch
            shard_paths: Paths to weight shard files
            dtype: Target dtype for conversion
        """
        if not self.config.enable_prefetch:
            return

        # Check if already prefetched/queued
        with self._lock:
            if component_name in self._cache:
                return

        item = PrefetchItem(
            component_name=component_name,
            shard_paths=shard_paths,
            dtype=dtype,
            device="cpu",  # Always prefetch to CPU, not GPU
        )

        try:
            self._queue.put_nowait(item)
            with self._lock:
                self._cache[component_name] = item

        except:
            # Queue full, skip prefetch
            pass

    def get(
        self,
        component_name: str,
        device: str,
        timeout: float = 60.0,
    ) -> Optional[Dict[str, torch.Tensor]]:
        """
        Get prefetched weights, transferring to GPU.

        If component is not prefetched, returns None.

        Args:
            component_name: Name of component
            device: Target GPU device
            timeout: Max wait time for prefetch completion

        Returns:
            Dict of tensors on GPU, or None if not prefetched
        """
        with self._lock:
            item = self._cache.get(component_name)

        if item is None:
            return None

        # Wait for prefetch to complete
        if not item.ready.wait(timeout=timeout):
            return None

        # Transfer from pinned CPU to GPU
        if item.tensors is not None:
            loader = self._get_loader()
            result = loader.transfer_to_device(
                item.tensors, device, non_blocking=True
            )

            # Clear from cache to free pinned memory
            with self._lock:
                del self._cache[component_name]

            return result

        return None

    def clear(self) -> None:
        """Clear all prefetched data."""
        with self._lock:
            self._cache.clear()
        # Drain queue
        while not self._queue.empty():
            try:
                self._queue.get_nowait()
            except Empty:
                break

    def _worker_loop(self) -> None:
        """Background worker that performs prefetch operations."""
        while self._running:
            try:
                item = self._queue.get(timeout=0.5)
            except Empty:
                continue

            try:
                # Load to pinned memory (NOT to GPU)
                loader = self._get_loader()
                item.tensors = loader.load_shards_parallel(
                    item.shard_paths,
                    to_pinned=True,
                    dtype=item.dtype,
                )

            except Exception as e:
                # Log error but don't crash — prefetch is best-effort
                import sys
                print(f"[PrefetchQueue] Error prefetching {item.component_name}: {e}", file=sys.stderr)
                item.tensors = None

            finally:
                item.ready.set()


# ============================================================================
# Global Instances
# ============================================================================

_prefetch_queue: Optional[PrefetchQueue] = None


def get_prefetch_queue() -> PrefetchQueue:
    """Get global PrefetchQueue instance."""
    global _prefetch_queue
    if _prefetch_queue is None:
        _prefetch_queue = PrefetchQueue()
        _prefetch_queue.start()
    return _prefetch_queue


def shutdown_io() -> None:
    """Shutdown I/O system (stop prefetch worker)."""
    global _prefetch_queue
    if _prefetch_queue is not None:
        _prefetch_queue.stop()
        _prefetch_queue = None



# core/prism/common/allocation.py
"""
Allocation Utilities for Prism

Extracted from solver.py and smart_solver.py.
Common allocation helpers used by the Prism solver.
"""

from typing import Dict, List, Tuple, Optional

# Use consolidated dtype module (eliminates code duplication)
from neurobrix.core.dtype import calculate_dtype_multiplier, get_dtype_bytes


def get_device_index(device_string: str) -> int:
    """
    Extract device index from device string.

    Args:
        device_string: Device string (e.g., 'cuda:2', 'cuda:0')

    Returns:
        Device index (e.g., 2, 0)
    """
    if ':' not in device_string:
        return 0
    return int(device_string.split(':')[1])


def allocate_sequential_shards(
    shards_with_sizes: List[Tuple[str, int, float]],
    device_order: List[str],
    device_capacity: Dict[str, float],
    device_usage: Dict[str, float],
    log_prefix: str = "[Prism]",
) -> Tuple[Dict[str, str], List[str], float]:
    """
    Allocate shards SEQUENTIALLY for pipeline parallelism.
    
    Pipeline requires blocks to be sequential on GPUs:
    - GPU0: blocks 0, 1, 2 (sequential)
    - GPU1: blocks 3, 4, 5 (sequential)
    NOT: GPU0: blocks 0, 2, 4 (interleaved - would break pipeline)
    
    Args:
        shards_with_sizes: List of (shard_path, block_idx, size_mb) tuples,
                          sorted by block index
        device_order: List of devices in order of preference
        device_capacity: {device: capacity_mb}
        device_usage: {device: current_usage_mb} - MODIFIED IN PLACE
        log_prefix: Prefix for log messages
        
    Returns:
        (shard_map, devices_used, total_memory_mb)
    """
    shard_map: Dict[str, str] = {}
    devices_used: List[str] = []
    total_memory_mb = 0.0
    
    current_device_idx = 0
    current_device = device_order[current_device_idx]
    
    for shard_path, block_idx, shard_mb in shards_with_sizes:
        # Check if current device has space
        free_space = device_capacity[current_device] - device_usage[current_device]
        
        # If not enough space, move to next device
        while shard_mb > free_space and current_device_idx < len(device_order) - 1:
            current_device_idx += 1
            current_device = device_order[current_device_idx]
            free_space = device_capacity[current_device] - device_usage[current_device]
            
        # Check if we can fit on current device
        if shard_mb <= free_space:
            assigned_device = current_device
            device_usage[current_device] += shard_mb
        else:
            # No GPU space, must go to CPU
            assigned_device = "cpu"
        shard_map[shard_path] = assigned_device
        total_memory_mb += shard_mb
        
        if assigned_device not in devices_used and assigned_device != "cpu":
            devices_used.append(assigned_device)
            
    return shard_map, devices_used, total_memory_mb





# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang

import warnings

import torch

from fla.ops.generalized_delta_rule.dplr.chunk_A_bwd import chunk_dplr_bwd_dqk_intra
from fla.ops.generalized_delta_rule.dplr.chunk_A_fwd import chunk_dplr_fwd_intra
from fla.ops.generalized_delta_rule.dplr.chunk_h_bwd import chunk_dplr_bwd_dhu
from fla.ops.generalized_delta_rule.dplr.chunk_h_fwd import chunk_dplr_fwd_h
from fla.ops.generalized_delta_rule.dplr.chunk_o_bwd import chunk_dplr_bwd_dAu, chunk_dplr_bwd_dv, chunk_dplr_bwd_o
from fla.ops.generalized_delta_rule.dplr.chunk_o_fwd import chunk_dplr_fwd_o
from fla.ops.generalized_delta_rule.dplr.wy_fast_bwd import chunk_dplr_bwd_wy
from fla.ops.generalized_delta_rule.dplr.wy_fast_fwd import prepare_wy_repr_fwd
from fla.ops.rwkv6.chunk import chunk_rwkv6_fwd_cumsum
from fla.utils import autocast_custom_bwd, autocast_custom_fwd, input_guard


def chunk_dplr_fwd(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    a: torch.Tensor,
    b: torch.Tensor,
    gk: torch.Tensor,
    scale: float,
    initial_state: torch.Tensor,
    output_final_state: bool,
    cu_seqlens: torch.LongTensor | None = None,
    chunk_size: int = 64,
):
    gi, ge = chunk_rwkv6_fwd_cumsum(gk, chunk_size, cu_seqlens=cu_seqlens)

    A_ab, A_qk, A_ak, A_qb, qg, kg, ag, bg = chunk_dplr_fwd_intra(
        q=q,
        k=k,
        a=a,
        b=b,
        gi=gi,
        ge=ge,
        scale=scale,
        cu_seqlens=cu_seqlens,
        chunk_size=chunk_size,
    )
    del ge

    # A_ab, A_ak, gi, ge torch.float32
    # A_qk, A_qb, qg, kg, ag, bg, dtype=q.dtype, eg: bf16
    w, u, _ = prepare_wy_repr_fwd(
        ag=ag,
        A_ab=A_ab,
        A_ak=A_ak,
        v=v,
        cu_seqlens=cu_seqlens,
        chunk_size=chunk_size,
    )
    del A_ab, A_ak
    h, v_new, final_state = chunk_dplr_fwd_h(
        kg=kg,
        bg=bg,
        v=v,
        w=w,
        u=u,
        gk=gi,
        initial_state=initial_state,
        output_final_state=output_final_state,
        cu_seqlens=cu_seqlens,
        chunk_size=chunk_size,
    )
    del u, kg, bg, gi

    o = chunk_dplr_fwd_o(
        qg=qg,
        v=v,
        v_new=v_new,
        A_qk=A_qk,
        A_qb=A_qb,
        h=h,
        cu_seqlens=cu_seqlens,
        chunk_size=chunk_size,
    )
    del v_new, h, A_qk, A_qb

    return o, final_state


class ChunkDPLRDeltaRuleFunction(torch.autograd.Function):

    @staticmethod
    @input_guard
    @autocast_custom_fwd
    def forward(
        ctx,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        a: torch.Tensor,
        b: torch.Tensor,
        gk: torch.Tensor,
        scale: float,
        initial_state: torch.Tensor,
        output_final_state: bool,
        cu_seqlens: torch.LongTensor | None = None,
    ):
        chunk_size = 16
        o, final_state = chunk_dplr_fwd(
            q=q,
            k=k,
            v=v,
            a=a,
            b=b,
            gk=gk,
            scale=scale,
            initial_state=initial_state,
            output_final_state=output_final_state,
            cu_seqlens=cu_seqlens,
            chunk_size=chunk_size,
        )
        ctx.save_for_backward(q, k, v, a, b, gk, initial_state)
        ctx.cu_seqlens = cu_seqlens
        ctx.scale = scale
        ctx.chunk_size = chunk_size
        return o.to(q.dtype), final_state

    @staticmethod
    @input_guard
    @autocast_custom_bwd
    def backward(
        ctx,
        do: torch.Tensor,
        dht: torch.Tensor,
    ):
        q, k, v, a, b, gk, initial_state = ctx.saved_tensors
        chunk_size = ctx.chunk_size
        cu_seqlens = ctx.cu_seqlens
        scale = ctx.scale

        # ******* start recomputing everything, otherwise i believe the gpu memory will be exhausted *******
        gi, ge = chunk_rwkv6_fwd_cumsum(gk, chunk_size, cu_seqlens=cu_seqlens)

        A_ab, A_qk, A_ak, A_qb, qg, kg, ag, bg = chunk_dplr_fwd_intra(
            q=q,
            k=k,
            a=a,
            b=b,
            gi=gi,
            ge=ge,
            scale=scale,
            cu_seqlens=cu_seqlens,
            chunk_size=chunk_size,
        )
        w, u, A_ab_inv = prepare_wy_repr_fwd(
            ag=ag,
            A_ab=A_ab,
            A_ak=A_ak,
            v=v,
            cu_seqlens=cu_seqlens,
            chunk_size=chunk_size,
        )
        del A_ab
        h, v_new, _ = chunk_dplr_fwd_h(
            kg=kg,
            bg=bg,
            v=v,
            w=w,
            u=u,
            gk=gi,
            initial_state=initial_state,
            cu_seqlens=cu_seqlens,
            chunk_size=chunk_size,
        )
        del u
        # ******* end of recomputation *******
        # A_ak, A_ab_inv, gi, ge torch.float32
        # A_qk, A_qb, qg, kg, ag, bg, v_new dtype=q.dtype, eg: bf16

        dv_new_intra, dA_qk, dA_qb = chunk_dplr_bwd_dAu(
            v=v,
            v_new=v_new,
            do=do,
            A_qb=A_qb,
            scale=scale,
            cu_seqlens=cu_seqlens,
            chunk_size=chunk_size,
        )

        dh, dh0, dv_new = chunk_dplr_bwd_dhu(
            qg=qg,
            bg=bg,
            w=w,
            gk=gi,
            h0=initial_state,
            dht=dht,
            do=do,
            dv=dv_new_intra,
            cu_seqlens=cu_seqlens,
            chunk_size=chunk_size,
        )

        dv = chunk_dplr_bwd_dv(
            A_qk=A_qk,
            kg=kg,
            do=do,
            dh=dh,
            cu_seqlens=cu_seqlens,
            chunk_size=chunk_size,
        )
        del A_qk

        dqg, dkg, dw, dbg, dgk_last = chunk_dplr_bwd_o(
            k=kg,
            b=bg,
            v=v,
            v_new=v_new,
            do=do,
            h=h,
            dh=dh,
            dv=dv_new,
            w=w,
            gk=gi,
            cu_seqlens=cu_seqlens,
            chunk_size=chunk_size,
            scale=scale,
        )
        del v_new

        dA_ab, dA_ak, dv, dag = chunk_dplr_bwd_wy(
            A_ab_inv=A_ab_inv,
            A_ak=A_ak,
            v=v,
            ag=ag,
            dw=dw,
            du=dv_new,
            dv0=dv,
            cu_seqlens=cu_seqlens,
            chunk_size=chunk_size,
        )
        del A_ak

        dq, dk, da, db, dgk = chunk_dplr_bwd_dqk_intra(
            q=q,
            k=k,
            a=a,
            b=b,
            gi=gi,
            ge=ge,
            dAqk=dA_qk,
            dAqb=dA_qb,
            dAak=dA_ak,
            dAab=dA_ab,
            dgk_last=dgk_last,
            dqg=dqg,
            dkg=dkg,
            dag=dag,
            dbg=dbg,
            chunk_size=chunk_size,
            scale=scale,
            cu_seqlens=cu_seqlens,
        )

        return dq.to(q), dk.to(k), dv.to(v), da.to(a), db.to(b), dgk.to(gk), None, dh0, None, None


@torch.compiler.disable
def chunk_dplr_delta_rule(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    a: torch.Tensor,
    b: torch.Tensor,
    gk: torch.Tensor,
    scale: float | None = None,
    initial_state: torch.Tensor | None = None,
    output_final_state: bool = False,
    cu_seqlens: torch.LongTensor | None = None,
    head_first: bool = False,
):
    r"""
    Args:
        q (torch.Tensor):
            queries of shape `[B, T, H, K]`.
        k (torch.Tensor):
            keys of shape `[B, T, H, K]`.
        v (torch.Tensor):
            values of shape `[B, T, H, V]`.
        a (torch.Tensor):
            activations of shape `[B, T, H, K]`.
        b (torch.Tensor):
            betas of shape `[B, T, H, K]`.
        gk (torch.Tensor):
            gk of shape `[B, T, H, K]`. decay term in log space!
        scale (Optional[float]):
            Scale factor for the RetNet attention scores.
            If not provided, it will default to `1 / sqrt(K)`. Default: `None`.
        initial_state (Optional[torch.Tensor]):
            Initial state of shape `[N, H, K, V]` for `N` input sequences.
            For equal-length input sequences, `N` equals the batch size `B`.
            Default: `None`.
        output_final_state (Optional[bool]):
            Whether to output the final state of shape `[N, H, K, V]`. Default: `False`.
        cu_seqlens (torch.LongTensor):
            Cumulative sequence lengths of shape `[N+1]` used for variable-length training,
            consistent with the FlashAttention API.
        head_first (Optional[bool]):
            Whether the inputs are in the head-first format. Default: `False`.
            This argument has been deprecated.

    Returns:
        o (torch.Tensor):
            Outputs of shape `[B, T, H, V]`.
        final_state (torch.Tensor):
            Final state of shape `[N, H, K, V]` if `output_final_state=True` else `None`.
    """
    if head_first:
        raise DeprecationWarning(
            "head_first is deprecated and will be removed in a future version. "
            "Please use head_first=False for now instead.",
        )
    if not head_first and q.shape[1] < q.shape[2]:
        warnings.warn(
            f"Input tensor shape suggests potential format mismatch: seq_len ({q.shape[1]}) < num_heads ({q.shape[2]}). "
            "This may indicate the inputs were passed in head-first format [B, H, T, ...] "
            "when head_first=False was specified. "
            "Please verify your input tensor format matches the expected shape [B, T, H, ...].",
        )
    if q.dtype == torch.float32:
        warnings.warn(
            """ChunkDeltaRuleFunction does not support float32 on some platforms. Please use bfloat16/float16.
            If you want to use float32, please solve the issue by yourself.""",
            category=RuntimeWarning,
            stacklevel=2,
        )
    if cu_seqlens is not None:
        if q.shape[0] != 1:
            raise ValueError(
                f"The batch size is expected to be 1 rather than {q.shape[0]} when using `cu_seqlens`."
                f"Please flatten variable-length inputs before processing.",
            )
        if initial_state is not None and initial_state.shape[0] != len(cu_seqlens) - 1:
            raise ValueError(
                f"The number of initial states is expected to be equal to the number of input sequences, "
                f"i.e., {len(cu_seqlens) - 1} rather than {initial_state.shape[0]}.",
            )
    scale = k.shape[-1] ** -0.5 if scale is None else scale
    o, final_state = ChunkDPLRDeltaRuleFunction.apply(
        q,
        k,
        v,
        a,
        b,
        gk,
        scale,
        initial_state,
        output_final_state,
        cu_seqlens,
    )
    return o, final_state



"""
Universal Gather - Triton Optimized (Auto-CUDA + Multidim Support + Scalar Fix)
"""
import torch
import triton
import triton.language as tl
from neurobrix.kernels.registry import register_kernel

# --- KERNEL 1 : EMBEDDING LOOKUP (Indices -> Vectors) ---
# Cas: Gather axis=0 sur une table [Vocab, Dim]
@triton.jit
def _gather_embedding_kernel(
    out_ptr, table_ptr, indices_ptr,
    stride_out_row, stride_table_row, stride_idx,
    dim, BLOCK_SIZE: tl.constexpr
):
    pid = tl.program_id(0) # Index dans le batch d'indices plats
    
    # 1. Lire l'index scalaire
    idx = tl.load(indices_ptr + pid * stride_idx)
    
    # 2. Copier la ligne correspondante de la table
    cols = tl.arange(0, BLOCK_SIZE)
    mask = cols < dim
    
    val = tl.load(table_ptr + idx * stride_table_row + cols, mask=mask, other=0.0)
    tl.store(out_ptr + pid * stride_out_row + cols, val, mask=mask)

# --- KERNEL 2 : GATHER STANDARD (Axis -1) ---
@triton.jit
def _gather_kernel(
    out_ptr, x_ptr, indices_ptr,
    stride_out_row, stride_x_row, stride_idx_row,
    n_cols, BLOCK_SIZE: tl.constexpr
):
    row_idx = tl.program_id(0)
    cols = tl.arange(0, BLOCK_SIZE)
    mask = cols < n_cols
    
    idx_val = tl.load(indices_ptr + row_idx * stride_idx_row + cols, mask=mask, other=0)
    val = tl.load(x_ptr + row_idx * stride_x_row + idx_val, mask=mask, other=0.0)
    
    tl.store(out_ptr + row_idx * stride_out_row + cols, val, mask=mask)

@triton.jit
def _gather_nd_kernel(
    x_ptr, indices_ptr, out_ptr,
    n_indices, index_dim,
    stride_x, stride_out,
    BLOCK_SIZE: tl.constexpr
):
    """General gather along specified dimension."""
    pid = tl.program_id(0)
    if pid >= n_indices:
        return

    # Load the index for this position
    idx = tl.load(indices_ptr + pid)

    # Copy the value
    for d_start in range(0, index_dim, BLOCK_SIZE):
        d_offs = d_start + tl.arange(0, BLOCK_SIZE)
        mask = d_offs < index_dim

        val = tl.load(x_ptr + idx * stride_x + d_offs, mask=mask, other=0.0)
        tl.store(out_ptr + pid * stride_out + d_offs, val, mask=mask)


def _gather_impl(
    x: torch.Tensor,
    indices: torch.Tensor,
    dim: int = 0,
) -> torch.Tensor:
    """Gather implementation using Triton."""
    assert x.is_cuda and indices.is_cuda
    x = x.contiguous()
    indices = indices.contiguous()

    ndim = x.ndim
    if dim < 0:
        dim = ndim + dim

    # For dim=0 gather (embedding-style), use the embedding kernel pattern
    if dim == 0 and x.ndim == 2:
        indices_flat = indices.flatten()
        num_indices = indices_flat.numel()
        embedding_dim = x.shape[1]

        out_flat = torch.empty(num_indices, embedding_dim, dtype=x.dtype, device=x.device)

        BLOCK_SIZE = min(1024, triton.next_power_of_2(embedding_dim))
        grid = (num_indices,)

        with torch.cuda.device(x.device):
            _gather_nd_kernel[grid](
                x, indices_flat, out_flat,
                num_indices, embedding_dim,
                x.stride(0), out_flat.stride(0),
                BLOCK_SIZE=BLOCK_SIZE,
                num_warps=4,
            )

        out_shape = list(indices.shape) + [embedding_dim]
        return out_flat.view(out_shape)

    # For gather along last dim
    if dim == ndim - 1:
        # Flatten to 2D
        orig_shape = list(x.shape)
        n_cols = orig_shape[-1]
        n_rows = 1
        for s in orig_shape[:-1]:
            n_rows *= s

        x_flat = x.view(n_rows, n_cols)
        indices_flat = indices.view(n_rows, -1)
        out_n_cols = indices_flat.shape[-1]

        out_flat = torch.empty(n_rows, out_n_cols, dtype=x.dtype, device=x.device)

        BLOCK_SIZE = min(1024, triton.next_power_of_2(out_n_cols))
        grid = (n_rows,)

        with torch.cuda.device(x.device):
            _gather_kernel[grid](
                out_flat, x_flat, indices_flat,
                out_flat.stride(0), x_flat.stride(0), indices_flat.stride(0),
                out_n_cols,
                BLOCK_SIZE=BLOCK_SIZE,
                num_warps=4,
            )

        out_shape = list(orig_shape[:-1]) + [out_n_cols]
        return out_flat.view(out_shape)

    # General case: move dim to position 0 and handle
    perm = [dim] + [i for i in range(ndim) if i != dim]
    x_t = x.permute(perm).contiguous()

    indices_shape = indices.shape
    indices_flat = indices.flatten()
    num_indices = indices_flat.numel()

    # After permute, shape is [dim_size, ...]
    inner_size = x_t[0].numel()

    out_flat = torch.empty(num_indices, inner_size, dtype=x.dtype, device=x.device)

    BLOCK_SIZE = min(1024, triton.next_power_of_2(inner_size))
    grid = (num_indices,)

    with torch.cuda.device(x.device):
        _gather_nd_kernel[grid](
            x_t.view(x_t.shape[0], -1), indices_flat, out_flat,
            num_indices, inner_size,
            x_t.view(x_t.shape[0], -1).stride(0), out_flat.stride(0),
            BLOCK_SIZE=BLOCK_SIZE,
            num_warps=4,
        )

    # Reshape and unpermute
    out_shape = list(indices_shape) + list(x_t.shape[1:])
    return out_flat.view(out_shape)


@register_kernel(family="llm", vendor="nvidia", tier="common", op_name="gather")
def gather_kernel(
    x: torch.Tensor,
    indices: torch.Tensor,
    dim: int = 0,
) -> torch.Tensor:
    """Gather operation - pure Triton."""
    return _gather_impl(x, indices, dim)




import logging

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger(__name__)


def heur_block_n(args):
    return triton.next_power_of_2(args["N"])


def heur_num_warps(args):
    if args["N"] <= 1024:
        return 1
    elif args["N"] <= 2048:
        return 4
    else:
        return 8


@libentry()
@triton.autotune(configs=runtime.get_tuned_config("log_softmax"), key=["M", "N"])
@triton.heuristics(
    {
        "BLOCK_N": heur_block_n,
        "num_warps": heur_num_warps,
    }
)
@triton.jit
def log_softmax_kernel(
    output_ptr,
    input_ptr,
    M,
    N,
    K,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    USE_K: tl.constexpr,
):
    pid_m = tle.program_id(0)
    pid_k = tle.program_id(1)
    m_offset = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    n_offset = tl.arange(0, BLOCK_N)
    offset = m_offset[:, None] * N * K + n_offset[None, :] * K
    if USE_K:
        offset += pid_k
    mask = m_offset[:, None] < M and n_offset[None, :] < N
    input_ptrs = input_ptr + offset
    inp = tl.load(input_ptrs, mask=mask, other=-float("inf")).to(tl.float32)
    row_minus_max = inp - tl.max(inp, axis=1)[:, None]
    numerator = tl.exp(row_minus_max)
    denominator = tl.sum(numerator, axis=1)[:, None]
    softmax_output = tl.log(numerator / denominator)
    output_ptrs = output_ptr + offset
    tl.store(output_ptrs, softmax_output, mask=mask)


@libentry()
@triton.autotune(configs=runtime.get_tuned_config("log_softmax"), key=["M", "N"])
@triton.heuristics(
    {
        "BLOCK_N": heur_block_n,
        "num_warps": heur_num_warps,
    }
)
@triton.jit
def log_softmax_backward_kernel(
    out_ptr,
    out_grad_ptr,
    in_grad_ptr,
    M,
    N,
    K,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    BLOCK_N_SPLIT: tl.constexpr,
):
    pid_m = tle.program_id(0)
    pid_k = tle.program_id(1)
    m_offset = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    n_split_offset = tl.arange(0, BLOCK_N_SPLIT)
    n_offset = tl.arange(0, BLOCK_N)
    all_offsets = m_offset[:, None] * N * K + n_offset[None, :] * K + pid_k
    out_grad_ptrs_all = out_grad_ptr + all_offsets
    all_mask = m_offset[:, None] < M and n_offset[None, :] < N
    out_grad_all = tl.load(out_grad_ptrs_all, mask=all_mask).to(tl.float32)
    scale = tl.sum(out_grad_all, 1)
    # use for loop to split N dim to reduce register cost
    for n in range(0, tl.cdiv(BLOCK_N, BLOCK_N_SPLIT)):
        offsets = (
            m_offset[:, None] * N * K
            + n_split_offset[None, :] * K
            + n * BLOCK_N_SPLIT * K
            + pid_k
        )
        mask = m_offset[:, None] < M and n_split_offset[None, :] + n * BLOCK_N_SPLIT < N
        out_ptrs = out_ptr + offsets
        out = tl.load(out_ptrs, mask=mask).to(tl.float32)
        exp_out = tl.exp(out.to(tl.float32))
        out_grad_ptrs = out_grad_ptr + offsets
        out_grad = tl.load(out_grad_ptrs, mask=mask).to(tl.float32)

        in_grad = out_grad - exp_out * scale[:, None]
        in_grad_ptrs = in_grad_ptr + offsets
        tl.store(in_grad_ptrs, in_grad, mask=mask)


def log_softmax(self, dim, half_to_float=False):
    logger.debug("METAX GEMS LOG_SOFTMAX")

    assert dim >= -self.ndim and dim < self.ndim, "Invalid dim"
    dim = dim % self.ndim
    M = 1
    N = self.shape[dim]
    for i in range(dim):
        M *= self.shape[i]
    inp = self.contiguous()
    if half_to_float:
        dtype = torch.float32
    else:
        dtype = self.dtype
    out = torch.empty_like(inp, dtype=dtype)
    K = inp.numel() // M // N
    USE_K = K != 1

    grid = lambda meta: (
        triton.cdiv(M, meta["BLOCK_M"]),
        K,
    )
    with torch_device_fn.device(inp.device):
        log_softmax_kernel[grid](
            out,
            inp,
            M,
            N,
            K,
            USE_K=USE_K,
        )
    return out


def log_softmax_backward(grad_output, output, dim, input_dtype):
    logger.debug("METAX GEMS LOG_SOFTMAX VJP")

    assert dim >= -output.ndim and dim < output.ndim, "Invalid dim"
    dim = dim % output.ndim
    M = 1
    N = output.shape[dim]
    for i in range(dim):
        M *= output.shape[i]

    grad_output = grad_output.contiguous()
    in_grad = torch.empty_like(output, dtype=input_dtype)
    K = output.numel() // M // N

    grid = lambda meta: (
        triton.cdiv(M, meta["BLOCK_M"]),
        K,
    )
    with torch_device_fn.device(in_grad.device):
        log_softmax_backward_kernel[grid](
            output,
            grad_output,
            in_grad,
            M,
            N,
            K,
            BLOCK_N_SPLIT=1024,
        )
    return in_grad
